# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Optional, Tuple

import torch
from torch import nn

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import coil_combination_method


class SERANetDC(nn.Module):
    """SERANet Data consistency block, as presented in [Huang2019]_.

    References
    ----------
    .. [Huang2019] Huang, Q., Chen, X., Metaxas, D., Nadar, M.S. (2019). Brain Segmentation from k-Space with
        End-to-End Recurrent Attention Network. In: , et al. Medical Image Computing and Computer Assisted
        Intervention – MICCAI 2019. Lecture Notes in Computer Science(), vol 11766. Springer, Cham.
        https://doi.org/10.1007/978-3-030-32248-9_31
    """

    def __init__(self, fft_centered: bool, fft_normalization: str, spatial_dims: Tuple[int, ...]):
        """Inits :class:`SERANetDC`.

        Parameters
        ----------
        fft_centered: bool
            Whether to center the fft.
        fft_normalization: str
            Normalization to apply to the fft.
        spatial_dims: Tuple[int, ...]
            Spatial dimensions.
        """
        super().__init__()
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims
        self.dc_weight = nn.Parameter(torch.ones(1))

    def forward(
        self,
        prediction: torch.Tensor,
        prev_prediction: torch.Tensor,
        reference_kspace: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of :class:`SERANetDC`.

        Parameters
        ----------
        prediction : torch.Tensor
            Prediction. Shape: (batch, channels, height, width, complex)
        prev_prediction : torch.Tensor
            Previous prediction. Shape: (batch, channels, height, width, complex)
        reference_kspace : torch.Tensor
            Reference k-space. Shape: (batch, channels, height, width, complex)
        mask : torch.Tensor
            Subsampling mask. Shape: (batch, channels, height, width, 1)

        Returns
        -------
        torch.Tensor
            Data consistency k-space. Shape: (batch, channels, height, width, complex)
        """
        prediction = fft2(prediction.float(), self.fft_centered, self.fft_normalization, self.spatial_dims).to(
            prediction
        )
        if prediction.dim() < reference_kspace.dim():
            prediction = prediction.unsqueeze(1)
        zero = torch.zeros_like(prediction)
        soft_dc = torch.where(mask.bool(), prediction - reference_kspace, zero) * self.dc_weight
        return prev_prediction - soft_dc - prediction


class SERANetReconstructionBlock(torch.nn.Module):
    """Reconstruction Model block for End-to-End Recurrent Attention Network, as presented in [Huang2019]_.

    This model applies a combination of soft data consistency with the input model as a regularizer. A series of these
    blocks can be stacked to form the full variational network.

    References
    ----------
    .. [Huang2019] Huang, Q., Chen, X., Metaxas, D., Nadar, M.S. (2019). Brain Segmentation from k-Space with
       End-to-End Recurrent Attention Network. In: , et al. Medical Image Computing and Computer Assisted
       Intervention – MICCAI 2019. Lecture Notes in Computer Science(), vol 11766. Springer, Cham.
       https://doi.org/10.1007/978-3-030-32248-9_31
    """

    def __init__(
        self,
        num_reconstruction_blocks: int,
        reconstruction_model: torch.nn.Module,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
        coil_combination_method: str = "SENSE",
    ):
        """Inits :class:`SERANetReconstructionBlock`.

        Parameters
        ----------
        num_reconstruction_blocks : int
            Number of reconstruction blocks.
        reconstruction_model : torch.nn.Module
            Reconstruction model.
        fft_centered : bool, optional
            Whether to center the fft. Default is ``False``.
        fft_normalization : str, optional
            The normalization of the fft. Default is ``"backward"``.
        spatial_dims : Tuple[int, int], optional
            The spatial dimensions of the data. Default is ``None``.
        coil_dim : int, optional
            The dimension of the coil dimension. Default is ``1``.
        coil_combination_method : str, optional
            The coil combination method. Default is ``"SENSE"``.
        """
        super().__init__()
        self.reconstruction_module = torch.nn.ModuleList(
            [reconstruction_model for _ in range(num_reconstruction_blocks)]
        )
        self.model_name = self.reconstruction_module[0].__class__.__name__.lower()
        if self.model_name == "modulelist":
            self.model_name = self.reconstruction_module[0][0].__class__.__name__.lower()

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim
        self.coil_combination_method = coil_combination_method

        self.reconstruction_module_dc = torch.nn.ModuleList(
            [
                SERANetDC(
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,  # type: ignore
                )
                for _ in range(num_reconstruction_blocks)
            ]
        )

    def forward(
        self,
        prediction: torch.Tensor,
        ref_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of :class:`SERANetReconstructionBlock`.

        Parameters
        ----------
        prediction : torch.Tensor
            Prediction. Shape: [batch, channels, height, width, 2]
        ref_kspace : torch.Tensor
            Reference k-space. Shape: [batch, channels, height, width, 2]
        sensitivity_maps : torch.Tensor
            Sensitivity maps. Shape: [batch, coils, height, width, 2]
        mask : torch.Tensor
            Subsampling mask. Shape: [batch, 1, height, width, 1]

        Returns
        -------
        torch.Tensor
            Reconstruction. Shape: [batch, height, width, 2]
        """
        pred_reconstruction = []
        prev_reconstruction = ref_kspace.clone()
        for recon_block, dc_block in zip(self.reconstruction_module, self.reconstruction_module_dc):
            reconstruction = self.step(recon_block, prediction, ref_kspace, sensitivity_maps, mask)
            reconstruction = dc_block(reconstruction, prev_reconstruction, ref_kspace, mask)
            prev_reconstruction = reconstruction
            reconstruction = ifft2(reconstruction, self.fft_centered, self.fft_normalization, self.spatial_dims)
            pred_reconstruction.append(reconstruction)
        return pred_reconstruction

    def step(
        self,
        block: torch.nn.Module,
        pred: torch.Tensor,
        ref_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Step of the reconstruction block.

        Parameters
        ----------
        block : torch.nn.Module
            The block to apply.
        pred : torch.Tensor
            Prediction. Shape: [batch, height, width, 2]
        ref_kspace : torch.Tensor
            Reference k-space. Shape: [batch, channels, height, width, 2]
        sensitivity_maps : torch.Tensor
            Sensitivity maps. Shape: [batch, coils, height, width, 2]
        mask : torch.Tensor
            Subsampling mask. Shape: [batch, 1, height, width, 1]

        Returns
        -------
        torch.Tensor
            Reconstruction. Shape: [batch, height, width, 2]
        """
        if self.model_name == "unet":
            reconstruction = block(pred).permute(0, 2, 3, 1)
            reconstruction = torch.view_as_real(
                reconstruction[..., 0].float() + 1j * reconstruction[..., 1].float()
            ).to(reconstruction)
        elif "cascadenet" in self.model_name:
            reconstruction = ref_kspace.clone()
            for cascade in block:
                reconstruction = cascade(reconstruction, ref_kspace, sensitivity_maps, mask)
            reconstruction = torch.view_as_complex(
                coil_combination_method(
                    ifft2(
                        reconstruction,
                        centered=self.fft_centered,
                        normalization=self.fft_normalization,
                        spatial_dims=self.spatial_dims,
                    ),
                    sensitivity_maps,
                    method=self.coil_combination_method,
                    dim=self.coil_dim,
                )
            )
        else:
            reconstruction = pred.clone()

        return reconstruction


class SERANetRecurrentBlock(torch.nn.Module):
    """RecurrentModel block for End-to-End Recurrent Attention Network, as presented in [1]_.

    This model applies a combination of soft data consistency with the input model as a regularizer.
    A series of these blocks can be stacked to form the full variational network.

    References
    ----------

    .. [1] Pramanik A, Wu X, Jacob M. Joint calibrationless reconstruction and segmentation of parallel MRI. arXiv
        preprint arXiv:2105.09220. 2021 May 19.
    """

    def __init__(
        self,
        num_iterations: int,
        attention_model: torch.nn.Module,
        unet_model: torch.nn.Module,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
    ):
        """Inits :class:`SERANetRecurrentBlock`.

        Parameters
        ----------
        num_iterations : int
            Number of iterations for the recurrent block.
        attention_model : torch.nn.Module
            Attention model.
        unet_model : torch.nn.Module
            Unet model.
        fft_centered : bool, optional
            Whether to center the fft. Default is ``False``.
        fft_normalization : str, optional
            The normalization of the fft. Default is ``"backward"``.
        spatial_dims : Tuple[int, int], optional
            The spatial dimensions of the data. Default is ``None``.
        coil_dim : int, optional
            The dimension of the coil dimension. Default is ``1``.
        """
        super().__init__()
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim

        self.num_iterations = num_iterations
        self.recurrent_module_unet = unet_model
        self.recurrent_module_attention = attention_model
        self.recurrent_module_dc = SERANetDC(
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,  # type: ignore
        )

    def forward(
        self,
        pred_reconstruction: torch.Tensor,
        pred_segmentation: torch.Tensor,
        ref_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,  # pylint: disable=unused-argument
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of :class:`SERANetRecurrentBlock`.

        Parameters
        ----------
        pred_reconstruction : torch.Tensor
            Prediction. Shape: [batch, height, width, 2]
        pred_segmentation : torch.Tensor
            Prediction. Shape: [batch, num_classes, height, width]
        ref_kspace : torch.Tensor
            Reference k-space. Shape: [batch, channels, height, width, 2]
        sensitivity_maps : torch.Tensor
            Sensitivity maps. Shape: [batch, coils, height, width, 2]
        mask : torch.Tensor
            Subsampling mask. Shape: [batch, 1, height, width, 1]

        Returns
        -------
        torch.Tensor
            Reconstruction. Shape: [batch, height, width, 2]
        """
        attention_map = pred_segmentation.clone()  # TODO: remove this
        prev_prediction = ref_kspace.clone()
        for _ in range(self.num_iterations):
            attention_map = self.chan_complex_to_last_dim(
                self.recurrent_module_attention(
                    self.complex_to_chan_dim(pred_reconstruction), attention_map * pred_segmentation
                )
            )
            attention_map = self.recurrent_module_dc(attention_map, prev_prediction, ref_kspace, mask)
            prev_prediction = attention_map
            attention_map = self.recurrent_module_unet(self.complex_to_chan_dim(attention_map))
        return attention_map

    @staticmethod
    def complex_to_chan_dim(x: torch.Tensor) -> torch.Tensor:
        """Convert the last dimension of the input to complex."""
        b, c, h, w, two = x.shape
        if two != 2:
            raise AssertionError
        return x.permute(0, 4, 1, 2, 3).reshape(b, 2 * c, h, w)

    @staticmethod
    def chan_complex_to_last_dim(x: torch.Tensor) -> torch.Tensor:
        """Convert the last dimension of the input to complex."""
        b, c2, h, w = x.shape
        if c2 % 2 != 0:
            raise AssertionError
        c = torch.div(c2, 2, rounding_mode="trunc")
        return x.view(b, 2, c, h, w).permute(0, 2, 3, 4, 1).contiguous()



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:
# https://github.com/Project-MONAI/MONAI/blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/patchembedding.py#L28

from typing import Sequence, Type, Union

import numpy as np
import torch
import torch.nn.functional as F
from einops.layers.torch import Rearrange
from torch import nn

from atommic.collections.segmentation.nn.vit_base.utils import trunc_normal_

SUPPORTED_EMBEDDING_TYPES = {"conv", "perceptron"}


class PatchEmbeddingBlock(nn.Module):
    """Implementation of a patch embedding block, as presented in [Dosovitskiy2020]_.

    References
    ----------
    .. [Dosovitskiy2020] Dosovitskiy A, Beyer L, Kolesnikov A, Weissenborn D, Zhai X, Unterthiner T, Dehghani M,
        Minderer M, Heigold G, Gelly S, Uszkoreit J. An image is worth 16x16 words: Transformers for image recognition
        at scale. arXiv preprint arXiv:2010.11929. 2020 Oct 22.

    .. note::
        This is a wrapper for monai implementation of PatchEmbeddingBlock. See: https://github.com/Project-MONAI/MONAI/
        blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/patchembedding.py#L28
    """

    def __init__(
        self,
        in_channels: int,
        img_size: Union[Sequence[int], int],
        patch_size: Union[Sequence[int], int],
        hidden_size: int,
        num_heads: int,
        pos_embed: str,
        dropout_rate: float = 0.0,
        spatial_dims: int = 3,
    ):
        """Inits :class:`PatchEmbeddingBlock`.

        Parameters
        ----------
        in_channels : int
            Dimension of input channels.
        img_size : Union[Sequence[int], int]
            Dimension of input image.
        patch_size : Union[Sequence[int], int]
            Dimension of patch size.
        hidden_size : int
            Dimension of hidden layer.
        num_heads : int
            Number of attention heads.
        pos_embed : str
            Position embedding layer type.
        dropout_rate : float, optional
            Faction of the input units to drop. Default is ``0.0``.
        spatial_dims : int, optional
            Number of spatial dimensions. Default is ``3``.
        """
        super().__init__()
        if not 0 <= dropout_rate <= 1:
            raise ValueError("dropout_rate should be between 0 and 1.")
        if hidden_size % num_heads != 0:
            raise ValueError("hidden size should be divisible by num_heads.")

        self.pos_embed = pos_embed

        for m, p in zip(img_size, patch_size):  # type: ignore
            m = list(m) if isinstance(m, tuple) else [m]  # type: ignore
            p = list(p) if isinstance(p, tuple) else [p]  # type: ignore
            m = np.array(m)
            p = np.array(p)
            if m < p:
                raise ValueError("patch_size should be smaller than img_size.")
            if self.pos_embed == "perceptron" and m % p != 0:
                raise ValueError("patch_size should be divisible by img_size for perceptron.")

        self.n_patches = np.prod([im_d // p_d for im_d, p_d in zip(img_size, patch_size)])  # type: ignore
        self.patch_dim = int(in_channels * np.prod(patch_size))

        self.patch_embeddings: nn.Module
        if self.pos_embed == "conv":
            if spatial_dims == 2:
                self.patch_embeddings = nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=hidden_size,
                    kernel_size=patch_size,
                    stride=patch_size,
                )
            elif spatial_dims == 3:
                self.patch_embeddings = nn.Conv3d(
                    in_channels=in_channels,
                    out_channels=hidden_size,
                    kernel_size=patch_size,
                    stride=patch_size,
                )
            else:
                raise ValueError(f"Convolutional patch embedding not supported for {spatial_dims}D.")
        elif self.pos_embed == "perceptron":
            # for 3d: "b c (h p1) (w p2) (d p3)-> b (h w d) (p1 p2 p3 c)"
            chars = (("h", "p1"), ("w", "p2"), ("d", "p3"))[:spatial_dims]
            from_chars = "b c " + " ".join(f"({k} {v})" for k, v in chars)
            to_chars = f"b ({' '.join([c[0] for c in chars])}) ({' '.join([c[1] for c in chars])} c)"
            axes_len = {f"p{i + 1}": p for i, p in enumerate(patch_size)}  # type: ignore
            self.patch_embeddings = nn.Sequential(
                Rearrange(f"{from_chars} -> {to_chars}", **axes_len),
                nn.Linear(self.patch_dim, hidden_size),
            )
        self.position_embeddings = nn.Parameter(torch.zeros(1, self.n_patches, hidden_size))
        self.dropout = nn.Dropout(dropout_rate)
        trunc_normal_(self.position_embeddings, mean=0.0, std=0.02, a=-2.0, b=2.0)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m: nn.Module):
        """Initialize the weights of the module."""
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, mean=0.0, std=0.02, a=-2.0, b=2.0)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`PatchEmbeddingBlock`."""
        x = self.patch_embeddings(x)
        if self.pos_embed == "conv":
            x = x.flatten(2).transpose(-1, -2)
        embeddings = x + self.position_embeddings
        embeddings = self.dropout(embeddings)
        return embeddings


class PatchEmbed(nn.Module):
    """Implementation of a patch embedding block, as presented in [Liu2021]_.

    Unlike ViT patch embedding block: (1) input is padded to satisfy window size requirements (2) normalized if
    specified (3) position embedding is not used.

    References
    ----------
    .. [Liu2021] Liu Z, Lin Y, Cao Y, Hu H, Wei Y, Zhang Z, Lin S, Guo B. Swin transformer: Hierarchical vision
        transformer using shifted windows. InProceedings of the IEEE/CVF International Conference on Computer Vision
        2021 (pp. 10012-10022).

    .. note::
        This is a wrapper for monai implementation of PatchEmbeddingBlock. See: https://github.com/Project-MONAI/MONAI/
        blob/c38d503a587f1779914bd071a1b2d66a6d9080c2/monai/networks/blocks/patchembedding.py#L28
    """

    def __init__(
        self,
        patch_size: Union[Sequence[int], int] = 2,
        in_chans: int = 1,
        embed_dim: int = 48,
        norm_layer: Type[nn.LayerNorm] = nn.InstanceNorm2d,  # pylint: disable=unused-argument
        spatial_dims: int = 3,
    ):
        """Inits :class:`PatchEmbed`.

        Parameters
        ----------
        patch_size : Union[Sequence[int], int]
            Dimension of patch size. Default is ``2``.
        in_chans : int
            Dimension of input channels. Default is ``1``.
        embed_dim : int
            Dimension of embedding. Default is ``48``.
        norm_layer : Type[nn.LayerNorm]
            Normalization layer. Default is ``nn.InstanceNorm2d``.
        spatial_dims : int, optional
            Number of spatial dimensions. Default is ``3``.
        """
        super().__init__()
        if spatial_dims not in (2, 3):
            raise ValueError("Spatial dimension should be 2 or 3.")

        patch_size = (patch_size,) * spatial_dims  # type: ignore
        self.patch_size = patch_size
        self.embed_dim = embed_dim
        if spatial_dims == 2:
            self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
        elif spatial_dims == 3:
            self.proj = nn.Conv3d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
        else:
            raise ValueError(f"Convolutional patch embedding not supported for {spatial_dims}D.")
        if spatial_dims == 2:
            self.norm = nn.InstanceNorm2d(embed_dim)
        elif spatial_dims == 3:
            self.norm = nn.InstanceNorm3d(embed_dim)
        else:
            self.norm = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`PatchEmbeddingBlock`."""
        x_shape = x.size()
        if len(x_shape) == 5:
            _, _, d, h, w = x_shape
            if w % self.patch_size[2] != 0:  # type: ignore
                x = F.pad(x, (0, self.patch_size[2] - w % self.patch_size[2]))  # type: ignore
            if h % self.patch_size[1] != 0:  # type: ignore
                x = F.pad(x, (0, 0, 0, self.patch_size[1] - h % self.patch_size[1]))  # type: ignore
            if d % self.patch_size[0] != 0:  # type: ignore
                x = F.pad(x, (0, 0, 0, 0, 0, self.patch_size[0] - d % self.patch_size[0]))  # type: ignore

        elif len(x_shape) == 4:
            _, _, h, w = x_shape
            if w % self.patch_size[1] != 0:  # type: ignore
                x = F.pad(x, (0, self.patch_size[1] - w % self.patch_size[1]))  # type: ignore
            if h % self.patch_size[0] != 0:  # type: ignore
                x = F.pad(x, (0, 0, 0, self.patch_size[0] - h % self.patch_size[0]))  # type: ignore

        x = self.proj(x)
        if self.norm is not None:
            x_shape = x.size()
            x = x.flatten(2).transpose(1, 2)
            x = self.norm(x)
            if len(x_shape) == 5:
                d, wh, ww = x_shape[2], x_shape[3], x_shape[4]
                x = x.transpose(1, 2).view(-1, self.embed_dim, d, wh, ww)
            elif len(x_shape) == 4:
                wh, ww = x_shape[2], x_shape[3]
                x = x.transpose(1, 2).view(-1, self.embed_dim, wh, ww)
        return x



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/config/hydra_runner.py

import functools
import os
import sys
from argparse import Namespace
from typing import Any, Callable, Optional

from hydra._internal.utils import _run_hydra, get_args_parser
from hydra.core.config_store import ConfigStore
from hydra.types import TaskFunction
from omegaconf import DictConfig, OmegaConf


def _get_gpu_name():
    """Get GPU name."""
    try:
        import pynvml  # pylint: disable=import-outside-toplevel
    except (ImportError, ModuleNotFoundError):
        return None

    pynvml.nvmlInit()
    handle = pynvml.nvmlDeviceGetHandleByIndex(0)
    cuda_capability, _ = pynvml.nvmlDeviceGetCudaComputeCapability(handle)
    pynvml.nvmlShutdown()
    if cuda_capability == 8:
        return "a100"
    if cuda_capability == 9:
        return "h100"
    return None


OmegaConf.register_new_resolver("gpu_name", _get_gpu_name)

# multiple interpolated values in the config
OmegaConf.register_new_resolver("multiply", lambda x, y: x * y)


def hydra_runner(
    config_path: Optional[str] = ".", config_name: Optional[str] = None, schema: Optional[Any] = None
) -> Callable[[TaskFunction], Any]:
    """Decorator used for passing the Config paths to main function. Optionally registers a schema used for
    validation/providing default values.

    Parameters
    ----------
    config_path : str, optional
        Path to the config file. Default is ``.``.
    config_name : str, optional
        Name of the config file. Default is ``None``.
    schema : Any, optional
        Schema used for validation/providing default values. Default is ``None``.

    Returns
    -------
    A decorator that passes the config paths to the main function.
    """

    def decorator(task_function: TaskFunction) -> Callable[[], None]:
        """Decorator that passes the config paths to the main function."""

        @functools.wraps(task_function)
        def wrapper(cfg_passthrough: Optional[DictConfig] = None) -> Any:
            """Wrapper that passes the config paths to the main function."""
            # Check it config was passed.
            if cfg_passthrough is not None:
                return task_function(cfg_passthrough)
            args = get_args_parser()

            # Parse arguments in order to retrieve overrides
            parsed_args: Namespace = args.parse_args()

            # Get overriding args in dot string format
            overrides = parsed_args.overrides  # type: list

            # Disable the creation of .hydra subdir
            # https://hydra.cc/docs/tutorials/basic/running_your_app/working_directory
            overrides.append("hydra.output_subdir=null")
            # Hydra logging outputs only to stdout (no log file).
            # https://hydra.cc/docs/configure_hydra/logging
            overrides.append("hydra/job_logging=stdout")

            # Set run.dir ONLY for ExpManager "compatibility" - to be removed.
            overrides.append("hydra.run.dir=.")

            # Check if user set the schema.
            if schema is not None:
                # Create config store.
                cs = ConfigStore.instance()

                # Get the correct ConfigStore "path name" to "inject" the schema.
                if parsed_args.config_name is not None:
                    path, name = os.path.split(parsed_args.config_name)
                    # Make sure the path is not set - as this will disable validation scheme.
                    if path != "":
                        sys.stderr.write(
                            "ERROR Cannot set config file path using `--config-name` when "
                            "using schema. Please set path using `--config-path` and file name using "
                            "`--config-name` separately.\n"
                        )
                        sys.exit(1)
                else:
                    name = config_name

                # Register the configuration as a node under the name in the group.
                cs.store(name=name, node=schema)  # group=group,

            # Wrap a callable object with name `parse_args`
            # This is to mimic the ArgParser.parse_args() API.
            class _argparse_wrapper:
                """Wrapper for ArgParser.parse_args()."""

                def __init__(self, arg_parser):
                    self.arg_parser = arg_parser
                    self._actions = arg_parser._actions

                @staticmethod
                def parse_args(args=None, namespace=None):  # pylint: disable=unused-argument
                    """Parse arguments."""
                    return parsed_args

                    # no return value from run_hydra() as it may sometime actually run the task_function
                    # multiple times (--multirun)

            _run_hydra(  # pylint: disable=no-value-for-parameter
                args_parser=_argparse_wrapper(args),
                task_function=task_function,
                config_path=config_path,
                config_name=config_name,
            )

        return wrapper

    return decorator



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/ndrplz/ConvLSTM_pytorch/blob/master/convlstm.py

from typing import Any, List, Optional, Tuple

import torch
from torch import Tensor, nn


class ConvLSTMCell(nn.Module):
    """A Convolutional Long Short-Term Memory (LSTM) cell."""

    def __init__(self, input_dim: int, hidden_dim: int, kernel_size: Tuple[int, int], bias: bool = True):
        """Inits :class:`ConvLSTMCell`.

        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: (int, int)
            Size of the convolutional kernel.
        bias: bool
            Whether to add the bias. Default is ``True``.
        """
        super().__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        self.kernel_size = kernel_size
        self.padding = kernel_size[0] // 2, kernel_size[1] // 2
        self.bias = bias

        self.conv = nn.Conv2d(
            in_channels=self.input_dim + self.hidden_dim,
            out_channels=4 * self.hidden_dim,
            kernel_size=self.kernel_size,
            padding=self.padding,
            bias=self.bias,
        )

    def forward(self, input_tensor: torch.Tensor, cur_state: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of :class:`ConvLSTMCell`.

        Parameters
        ----------
        input_tensor: torch.Tensor
            Input tensor. Shape [batch_size, input_dim, height, width]
        cur_state: torch.Tensor
            Current state of the hidden state. Shape [batch_size, hidden_dim, height, width]

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            Tuple of the next hidden state and the cell state. Shape [batch_size, hidden_dim, height, width]
        """
        h_cur, c_cur = cur_state

        # concatenate along channel axis
        combined = torch.cat([input_tensor, h_cur], dim=1)

        combined_conv = self.conv(combined)
        cc_i, cc_f, cc_o, cc_g = torch.split(combined_conv, self.hidden_dim, dim=1)
        i = torch.sigmoid(cc_i)
        f = torch.sigmoid(cc_f)
        o = torch.sigmoid(cc_o)
        g = torch.tanh(cc_g)

        c_next = f * c_cur + i * g
        h_next = o * torch.tanh(c_next)

        return h_next, c_next

    def init_hidden(self, batch_size: int, image_size: Tuple[int, int]) -> Tuple[torch.Tensor, torch.Tensor]:
        """Initializes the hidden state.

        Parameters
        ----------
        batch_size: int
            Batch size.
        image_size: Tuple[int, int]
            Size of the image. Shape [height, width]

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            Tuple of the next hidden state and the cell state. Shape [batch_size, hidden_dim, height, width]
        """
        height, width = image_size
        return (
            torch.zeros(batch_size, self.hidden_dim, height, width, device=self.conv.weight.device),
            torch.zeros(batch_size, self.hidden_dim, height, width, device=self.conv.weight.device),
        )


class ConvLSTM(nn.Module):
    """A Convolutional Long Short-Term Memory (LSTM) block."""

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        kernel_size: Any,
        num_layers: int,
        batch_first: bool = False,
        bias: bool = True,
        return_all_layers: bool = False,
    ):
        """Inits :class:`ConvLSTM`.

        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: Any
            Size of the convolutional kernel.
        num_layers: int
            Number of layers.
        batch_first: bool
            Whether the first dimension corresponds to the batch size. Default is ``False``.
        bias: bool
            Whether to add the bias. Default is ``True``.
        return_all_layers: bool
            Whether to return all layers or just the last layer. Default is ``False``.
        """
        super().__init__()

        kernel_size = (kernel_size, kernel_size)

        self._check_kernel_size_consistency(kernel_size)

        # Make sure that both `kernel_size` and `hidden_dim` are lists having len == num_layers
        kernel_size = self._extend_for_multilayer(kernel_size, num_layers)
        hidden_dim = self._extend_for_multilayer(hidden_dim, num_layers)
        if not len(kernel_size) == len(hidden_dim) == num_layers:  # type: ignore
            raise ValueError("Inconsistent list length.")

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.kernel_size = kernel_size
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.bias = bias
        self.return_all_layers = return_all_layers

        cell_list = []
        for i in range(0, self.num_layers):
            cur_input_dim = self.input_dim if i == 0 else self.hidden_dim[i - 1]  # type: ignore
            cell_list.append(
                ConvLSTMCell(
                    input_dim=cur_input_dim,
                    hidden_dim=self.hidden_dim[i],  # type: ignore
                    kernel_size=self.kernel_size[i],
                    bias=self.bias,
                )
            )

        self.cell_list = nn.ModuleList(cell_list)

    def forward(
        self, input_tensor: torch.Tensor, hidden_state: Optional[List[Tuple[torch.Tensor, torch.Tensor]]] = None
    ) -> tuple[list[Tensor], list[list[Any]]]:
        """Forward pass of :class:`ConvLSTM`.

        Parameters
        ----------
        input_tensor : torch.Tensor
            Input tensor of shape [batch_size, seq_len, channels, height, width] if ``batch_first`` is ``False``.
            Otherwise, the shape should be [seq_len, batch_size, channels, height, width].
        hidden_state : Optional[List[Tuple[torch.Tensor, torch.Tensor]]]
            List of tuples of the hidden state and the cell state for each layer. The shape of each tensor should be
            [batch_size, hidden_dim, height, width]. If ``None``, the hidden state will be initialized to zero.

        Returns
        -------
        Tuple[torch.Tensor, List[Tuple[torch.Tensor, torch.Tensor]]]
            Tuple of the output tensor and the list of the hidden state and the cell state for each layer. The shape
            of the output tensor is [batch_size, seq_len, hidden_dim, height, width] if ``batch_first`` is ``False``.
            Otherwise, the shape should be [seq_len, batch_size, hidden_dim, height, width]. The shape of each tensor
            in the list is [batch_size, hidden_dim, height, width].
        """
        if not self.batch_first:
            # (t, b, c, h, w) -> (b, t, c, h, w)
            input_tensor = input_tensor.permute(1, 0, 2, 3, 4)

        b, _, _, h, w = input_tensor.size()

        # Implement stateful ConvLSTM
        if hidden_state is not None:
            raise NotImplementedError()

        # Since the init is done in forward. Can send image size here
        hidden_state = self._init_hidden(batch_size=b, image_size=(h, w))

        layer_output_list = []
        last_state_list = []

        seq_len = input_tensor.size(1)
        cur_layer_input = input_tensor

        for layer_idx in range(self.num_layers):
            h, c = hidden_state[layer_idx]
            output_inner = []
            for t in range(seq_len):
                h, c = self.cell_list[layer_idx](input_tensor=cur_layer_input[:, t, :, :, :], cur_state=[h, c])
                output_inner.append(h)

            layer_output = torch.stack(output_inner, dim=1)
            cur_layer_input = layer_output

            layer_output_list.append(layer_output)
            last_state_list.append([h, c])

        if not self.return_all_layers:
            layer_output_list = layer_output_list[-1:]
            last_state_list = last_state_list[-1:]

        return layer_output_list, last_state_list

    def _init_hidden(self, batch_size, image_size):
        """Initialize the hidden state."""
        init_states = []
        for i in range(self.num_layers):
            init_states.append(self.cell_list[i].init_hidden(batch_size, image_size))
        return init_states

    @staticmethod
    def _check_kernel_size_consistency(kernel_size):
        """Check the kernel size consistency."""
        if not (
            isinstance(kernel_size, tuple)
            or (isinstance(kernel_size, list) and all(isinstance(elem, tuple) for elem in kernel_size))
        ):
            raise ValueError("`kernel_size` must be tuple or list of tuples")

    @staticmethod
    def _extend_for_multilayer(param, num_layers):
        """Extend parameter for multilayer."""
        if not isinstance(param, list):
            param = [param] * num_layers
        return param



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.utils.atommic_logging import Logger as _Logger

logging = _Logger()
try:
    from atommic.utils.lightning_logger_patch import add_memory_handlers_to_pl_logger

    add_memory_handlers_to_pl_logger()
except ModuleNotFoundError:
    pass



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Sequence, Union

import numpy as np
import torch
from omegaconf import ListConfig

__all__ = ["fft2", "ifft2", "fftshift", "ifftshift"]


def fft2(
    x: torch.Tensor,
    centered: bool = False,
    normalization: str = "backward",
    spatial_dims: Sequence[int] = None,
) -> torch.Tensor:
    r"""Apply 2-dimensional Fast Fourier Transform.

    Parameters
    ----------
    x : torch.Tensor
        Complex valued input data.
    centered : bool
        Whether to center the fft. If True, the fft will be shifted so that the zero frequency component is in the
        center of the spectrum. Default is ``False``.
    normalization : str
        Normalization mode. For the forward transform (fft2()), these correspond to: \n
            * ``forward`` - normalize by 1/n
            * ``backward`` - no normalization
            * ``ortho`` - normalize by 1/sqrt(n) (making the FFT orthonormal)

        Where n = prod(s) is the logical FFT size.
        Calling the backward transform (ifft2()) with the same normalization mode will apply an overall
        normalization of 1/n between the two transforms. This is required to make ifft2() the exact inverse.
        Default is ``backward`` (no normalization).
    spatial_dims : Sequence[int]
        Dimensions to apply the FFT. Default is the last two dimensions. If tensor is viewed as real, the last
        dimension is assumed to be the complex dimension.

    Returns
    -------
    torch.Tensor
        The 2D FFT of the input.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.fft import fft2
    >>> data = torch.randn(2, 3, 4, 5, 2)
    >>> fft2(data).shape
    torch.Size([2, 3, 4, 5, 2])
    >>> fft2(data, centered=True, normalization="ortho", spatial_dims=[-3, -2]).shape
    torch.Size([2, 3, 4, 5, 2])

    .. note::
        The PyTorch fft2 function does not support complex tensors. Therefore, the input is converted to a complex
        tensor and then converted back to a real tensor. This is done by using the torch.view_as_complex and
        torch.view_as_real functions. The input is assumed to be a real tensor with the last dimension being the
        complex dimension.

        The PyTorch fft2 function performs a separate fft, so fft2 is the same as fft(fft(data, dim=-2), dim=-1).

        Source: https://pytorch.org/docs/stable/fft.html#torch.fft.fft2
    """
    if x.shape[-1] == 2:
        x = torch.view_as_complex(x)

    if spatial_dims is None:
        spatial_dims = [-2, -1]
    elif isinstance(spatial_dims, ListConfig):
        spatial_dims = list(spatial_dims)

    if centered:
        x = ifftshift(x, dim=spatial_dims)

    x = torch.fft.fft2(
        x,
        dim=spatial_dims,
        norm=normalization if normalization.lower() != "none" else None,
    )

    if centered:
        x = fftshift(x, dim=spatial_dims)

    x = torch.view_as_real(x)

    return x


def ifft2(
    x: torch.Tensor,
    centered: bool = False,
    normalization: str = "backward",
    spatial_dims: Sequence[int] = None,
) -> torch.Tensor:
    r"""Apply 2-dimensional Inverse Fast Fourier Transform.

    Parameters
    ----------
    x : torch.Tensor
        Complex valued input data.
    centered : bool
        Whether to center the ifft. If True, the ifft will be shifted so that the zero frequency component is in the
        center of the spectrum. Default is ``False``.
    normalization : str
        Normalization mode. For the backward transform (ifft2()), these correspond to: \n
            * ``forward`` - normalize by 1/n
            * ``backward`` - no normalization
            * ``ortho`` - normalize by 1/sqrt(n) (making the IFFT orthonormal)

        Where n = prod(s) is the logical IFFT size.
        Calling the forward transform (fft2()) with the same normalization mode will apply an overall
        normalization of 1/n between the two transforms. This is required to make fft2() the exact inverse.
        Default is ``backward`` (no normalization).
    spatial_dims : Sequence[int]
        Dimensions to apply the IFFT. Default is the last two dimensions. If tensor is viewed as real, the last
        dimension is assumed to be the complex dimension.

    Returns
    -------
    torch.Tensor
        The 2D IFFT of the input.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.fft import ifft2
    >>> data = torch.randn(2, 3, 4, 5, 2)
    >>> ifft2(data).shape
    torch.Size([2, 3, 4, 5, 2])
    >>> ifft2(data, centered=True, normalization="ortho", spatial_dims=[-3, -2]).shape
    torch.Size([2, 3, 4, 5, 2])

    .. note::
        The PyTorch ifft2 function does not support complex tensors. Therefore, the input is converted to a complex
        tensor and then converted back to a real tensor. This is done by using the torch.view_as_complex and
        torch.view_as_real functions. The input is assumed to be a real tensor with the last dimension being the
        complex dimension.

        The PyTorch ifft2 function performs a separate ifft, so ifft2 is the same as ifft(ifft(data, dim=-2), dim=-1).

        Source: https://pytorch.org/docs/stable/fft.html#torch.fft.ifft2
    """
    if x.shape[-1] == 2:
        x = torch.view_as_complex(x)

    if spatial_dims is None:
        spatial_dims = [-2, -1]
    elif isinstance(spatial_dims, ListConfig):
        spatial_dims = list(spatial_dims)

    if centered:
        x = ifftshift(x, dim=spatial_dims)

    x = torch.fft.ifft2(
        x,
        dim=spatial_dims,
        norm=normalization if normalization.lower() != "none" else None,
    )

    if centered:
        x = fftshift(x, dim=spatial_dims)

    x = torch.view_as_real(x)

    return x


def roll_one_dim(x: torch.Tensor, shift: int, dim: int) -> torch.Tensor:
    """Similar to roll but for only one dim.

    Parameters
    ----------
    x : torch.Tensor
        Input data.
    shift : int
        Amount to roll.
    dim : int
        Which dimension to roll.

    Returns
    -------
    torch.Tensor
        The rolled tensor.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.fft import roll_one_dim
    >>> data = torch.randn(2, 3, 4, 5)
    >>> roll_one_dim(data, 1, 0).shape
    torch.Size([2, 3, 4, 5])

    .. note::
        Source: https://github.com/facebookresearch/fastMRI/blob/main/fastmri/fftc.py
    """
    shift %= x.size(dim)
    if shift == 0:
        return x

    left = x.narrow(dim, 0, x.size(dim) - shift)
    right = x.narrow(dim, x.size(dim) - shift, shift)

    return torch.cat((right, left), dim=dim)


def roll(x: torch.Tensor, shift: List[int], dim: Union[List[int], Sequence[int]]) -> torch.Tensor:
    """Similar to np.roll but applies to PyTorch Tensors.

    Parameters
    ----------
    x : torch.Tensor
        Input data.
    shift : List[int]
        Amount to roll.
    dim : Union[List[int], Sequence[int]]
        Which dimension to roll.

    Returns
    -------
    torch.Tensor
        The rolled tensor.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.fft import roll
    >>> data = torch.randn(2, 3, 4, 5)
    >>> roll(data, [1, 2], [0, 1]).shape
    torch.Size([2, 3, 4, 5])

    .. note::
        Source: https://github.com/facebookresearch/fastMRI/blob/main/fastmri/fftc.py
    """
    if len(shift) != len(dim):
        raise ValueError("len(shift) must match len(dim)")

    if isinstance(dim, ListConfig):
        dim = list(dim)

    for s, d in zip(shift, dim):
        x = roll_one_dim(x, s, d)

    return x


def fftshift(x: torch.Tensor, dim: Union[List[int], Sequence[int]] = None) -> torch.Tensor:
    """Similar to np.fft.fftshift but applies to PyTorch Tensors.

    Parameters
    ----------
    x : torch.Tensor
        Input data.
    dim : Union[List[int], Sequence[int]]
        Which dimension to shift.

    Returns
    -------
    torch.Tensor
        The shifted tensor.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.fft import fftshift
    >>> data = torch.randn(2, 3, 4, 5)
    >>> fftshift(data).shape
    torch.Size([2, 3, 4, 5])

    .. note::
        Source: https://github.com/facebookresearch/fastMRI/blob/main/fastmri/fftc.py
    """
    if dim is None:
        # this weird code is necessary for torch.jit.script typing
        dim = [0] * (x.dim())
        for i in range(1, x.dim()):
            dim[i] = i
    elif isinstance(dim, ListConfig):
        dim = list(dim)
    elif not isinstance(dim, list):
        dim = [dim]  # type: ignore

    # Also necessary for torch.jit.script
    shift = [0] * len(dim)
    for i, dim_num in enumerate(dim):
        shift[i] = np.floor_divide(x.shape[dim_num], 2)

    return roll(x, shift, dim)


def ifftshift(x: torch.Tensor, dim: Union[List[int], Sequence[int]] = None) -> torch.Tensor:
    """Similar to np.fft.ifftshift but applies to PyTorch Tensors.

    Parameters
    ----------
    x : torch.Tensor
        Input data.
    dim : Union[List[int], Sequence[int]]
        Which dimension to shift.

    Returns
    -------
    torch.Tensor
        The shifted tensor.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.parts.fft import ifftshift
    >>> data = torch.randn(2, 3, 4, 5)
    >>> ifftshift(data).shape
    torch.Size([2, 3, 4, 5])

    .. note::
        Source: https://github.com/facebookresearch/fastMRI/blob/main/fastmri/fftc.py
    """
    if dim is None:
        # this weird code is necessary for torch.jit.script typing
        dim = [0] * (x.dim())
        for i in range(1, x.dim()):
            dim[i] = i
    elif isinstance(dim, ListConfig):
        dim = list(dim)
    elif not isinstance(dim, list):
        dim = [dim]  # type: ignore

    # Also necessary for torch.jit.script
    shift = [0] * len(dim)
    for i, dim_num in enumerate(dim):
        shift[i] = np.floor_divide(x.shape[dim_num] + 1, 2)

    return roll(x, shift, dim)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import Dict, List, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import fft2
from atommic.collections.common.parts.utils import check_stacked_complex, expand_op
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.rim_base.rim_block import RIMBlock
from atommic.core.classes.common import typecheck

__all__ = ["CIRIM"]


class CIRIM(BaseMRIReconstructionModel):
    """Implementation of the Cascades of Independently Recurrent Inference Machines, as presented in
    [Karkalousos2022]_.

    References
    ----------
    .. [Karkalousos2022] Karkalousos D, Noteboom S, Hulst HE, Vos FM, Caan MWA. Assessment of data consistency through
        cascades of independently recurrent inference machines for fast and robust accelerated MRI reconstruction.
        Phys Med Biol. 2022 Jun 8;67(12). doi: 10.1088/1361-6560/ac6cc2. PMID: 35508147.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`CIRIM`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        # make time-steps size divisible by 8 for fast fp16 training
        self.time_steps = 8 * math.ceil(cfg_dict.get("time_steps") / 8)
        self.no_dc = cfg_dict.get("no_dc")
        self.reconstruction_module = torch.nn.ModuleList(
            [
                RIMBlock(
                    recurrent_layer=cfg_dict.get("recurrent_layer"),
                    conv_filters=cfg_dict.get("conv_filters"),
                    conv_kernels=cfg_dict.get("conv_kernels"),
                    conv_dilations=cfg_dict.get("conv_dilations"),
                    conv_bias=cfg_dict.get("conv_bias"),
                    recurrent_filters=cfg_dict.get("recurrent_filters"),
                    recurrent_kernels=cfg_dict.get("recurrent_kernels"),
                    recurrent_dilations=cfg_dict.get("recurrent_dilations"),
                    recurrent_bias=cfg_dict.get("recurrent_bias"),
                    depth=cfg_dict.get("depth"),
                    time_steps=self.time_steps,
                    conv_dim=cfg_dict.get("conv_dim"),
                    no_dc=self.no_dc,
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim,
                    dimensionality=cfg_dict.get("dimensionality"),
                    coil_combination_method=self.coil_combination_method,
                )
                for _ in range(cfg_dict.get("num_cascades"))
            ]
        )

        # Keep estimation through the cascades if keep_prediction is True or re-estimate it if False.
        self.keep_prediction = cfg_dict.get("keep_prediction")

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,
    ) -> Union[List[List[torch.Tensor]], List[torch.Tensor], torch.Tensor]:
        """Forward pass of :class:`CIRIM`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        List of torch.Tensor
            List of the intermediate predictions for each cascade. Shape [batch_size, n_x, n_y].
        """
        prediction = y.clone()
        initial_prediction = None if initial_prediction is None or initial_prediction.dim() < 4 else initial_prediction
        hx = None
        cascades_predictions = []
        for i, cascade in enumerate(self.reconstruction_module):
            # Forward pass through the cascades
            prediction, hx = cascade(
                prediction,
                y,
                sensitivity_maps,
                mask,
                initial_prediction if i == 0 else prediction,
                hx,
                sigma,
                keep_prediction=False if i == 0 else self.keep_prediction,
            )
            cascades_predictions.append([check_stacked_complex(p) for p in prediction])
            prediction = prediction[-1]
        return cascades_predictions

    def process_reconstruction_loss(  # noqa: MC0001
        self,
        target: torch.Tensor,
        prediction: Union[List[List[torch.Tensor]], List[torch.Tensor], torch.Tensor],
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        attrs: Union[Dict, torch.Tensor],
        r: Union[int, torch.Tensor],
        loss_func: torch.nn.Module,
    ) -> torch.Tensor:
        """Processes the reconstruction loss for the CIRIM model. It differs from the base class in that it can handle
        multiple cascades and time steps.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, n_x, n_y, 2].
        prediction : Union[list, torch.Tensor]
            Prediction(s) of shape [batch_size, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2]. It will be used if self.ssdu is True, to
            expand the target and prediction to multiple coils.
        mask : torch.Tensor
            Mask of shape [batch_size, n_x, n_y, 2]. It will be used if self.ssdu is True, to enforce data consistency
            on the prediction.
        attrs : Dict
            Attributes of the data with pre normalization values.
        r : int
            The selected acceleration factor.
        loss_func : torch.nn.Module
            Loss function. Must be one of {torch.nn.L1Loss(), torch.nn.MSELoss(),
            atommic.collections.reconstruction.losses.ssim.SSIMLoss()}. Default is ``torch.nn.L1Loss()``.

        Returns
        -------
        loss: torch.FloatTensor
            If self.accumulate_loss is True, returns an accumulative result of all intermediate losses.
            Otherwise, returns the loss of the last intermediate loss.
        """
        # If kspace reconstruction loss is used, the target needs to be transformed to k-space.
        if self.kspace_reconstruction_loss:
            # If inputs are complex, then they need to be viewed as real.
            if target.shape[-1] != 2 and torch.is_complex(target):
                target = torch.view_as_real(target)
            # If SSDU is used, then the coil-combined inputs need to be expanded to multiple coils using the
            # sensitivity maps.
            if self.ssdu:
                target = expand_op(target, sensitivity_maps, self.coil_dim)
            # Transform to k-space.
            target = fft2(target, self.fft_centered, self.fft_normalization, self.spatial_dims)
            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target)
        elif not self.unnormalize_loss_inputs:
            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target)
            # Normalize inputs to [0, 1]
            target = torch.abs(target / torch.max(torch.abs(target)))

        def compute_reconstruction_loss(t, p, s):
            if self.unnormalize_loss_inputs:
                # we do the unnormalization here to avoid explicitly iterating through list of predictions, which
                # might be a list of lists.
                t, p, s = self.__unnormalize_for_loss_or_log__(t, p, s, attrs, r)

            # If kspace reconstruction loss is used, the target needs to be transformed to k-space.
            if self.kspace_reconstruction_loss:
                # If inputs are complex, then they need to be viewed as real.
                if p.shape[-1] != 2 and torch.is_complex(p):
                    p = torch.view_as_real(p)
                # If SSDU is used, then the coil-combined inputs need to be expanded to multiple coils using the
                # sensitivity maps.
                if self.ssdu:
                    p = expand_op(p, s, self.coil_dim)
                # Transform to k-space.
                p = fft2(p, self.fft_centered, self.fft_normalization, self.spatial_dims)
                # If SSDU is used, then apply the mask to the prediction to enforce data consistency.
                if self.ssdu:
                    p = p * mask
                # Ensure loss inputs are both viewed in the same way.
                p = self.__abs_output__(p)
            elif not self.unnormalize_loss_inputs:
                p = self.__abs_output__(p)
                # Normalize inputs to [0, 1]
                p = torch.abs(p / torch.max(torch.abs(p)))

            if "ssim" in str(loss_func).lower():
                return loss_func(
                    t.unsqueeze(dim=self.coil_dim),
                    p.unsqueeze(dim=self.coil_dim),
                    data_range=torch.tensor(
                        [max(torch.max(t).item(), torch.max(p).item()) - min(torch.min(t).item(), torch.min(p).item())]
                    )
                    .unsqueeze(dim=0)
                    .to(t.device),
                )

            return loss_func(t, p)

        if self.accumulate_predictions:
            cascades_weights = torch.logspace(-1, 0, steps=len(prediction)).to(target.device)
            cascades_loss = []
            for cascade_pred in prediction:
                time_steps_weights = torch.logspace(-1, 0, steps=self.time_steps).to(target.device)
                if self.num_echoes > 0:
                    time_steps_loss = [
                        torch.mean(
                            torch.stack(
                                [
                                    compute_reconstruction_loss(
                                        target[echo].unsqueeze(0), time_step_pred[echo].unsqueeze(0), sensitivity_maps
                                    )
                                    for echo in range(target.shape[0])
                                ]
                            )
                        ).to(target)
                        for time_step_pred in cascade_pred
                    ]
                else:
                    time_steps_loss = [
                        compute_reconstruction_loss(target, time_step_pred, sensitivity_maps)
                        for time_step_pred in cascade_pred
                    ]
                cascade_loss = sum(x * w for x, w in zip(time_steps_loss, time_steps_weights)) / self.time_steps
                cascades_loss.append(cascade_loss)
            loss = sum(x * w for x, w in zip(cascades_loss, cascades_weights)) / len(prediction)
        else:
            # keep the last prediction of the last cascade
            prediction = prediction[-1][-1]
            loss = compute_reconstruction_loss(target, prediction, sensitivity_maps)

        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import os
import re
import warnings
from pathlib import Path
from typing import Callable, Optional, Tuple, Union

import h5py
import numpy as np

from atommic.collections.common.data.mri_loader import MRIDataset
from atommic.collections.common.parts.utils import is_none


class qMRIDataset(MRIDataset):
    """A dataset class for quantitative MRI.

    .. note::
        Extends :class:`atommic.collections.common.data.MRIDataset`.
    """

    def __init__(
        self,
        root: Union[str, Path, os.PathLike],
        coil_sensitivity_maps_root: Union[str, Path, os.PathLike] = None,
        mask_root: Union[str, Path, os.PathLike] = None,
        noise_root: Union[str, Path, os.PathLike] = None,
        initial_predictions_root: Union[str, Path, os.PathLike] = None,
        dataset_format: str = None,
        sample_rate: Optional[float] = None,
        volume_sample_rate: Optional[float] = None,
        use_dataset_cache: bool = False,
        dataset_cache_file: Union[str, Path, os.PathLike] = None,
        num_cols: Optional[Tuple[int]] = None,
        consecutive_slices: int = 1,
        data_saved_per_slice: bool = False,
        n2r_supervised_rate: Optional[float] = 0.0,
        complex_target: bool = False,
        log_images_rate: Optional[float] = 1.0,
        transform: Optional[Callable] = None,
        sequence: str = None,
        segmentation_mask_root: Union[str, Path, os.PathLike] = None,
        kspace_scaling_factor: float = 1.0,
        **kwargs,
    ):
        """Inits :class:`qMRIDataset`.

        Parameters
        ----------
        root : Union[str, Path, os.PathLike]
            Path to the dataset.
        sense_root : Union[str, Path, os.PathLike], optional
            Path to the coil sensitivities maps dataset, if stored separately.
        mask_root : Union[str, Path, os.PathLike], optional
            Path to stored masks, if stored separately.
        noise_root : Union[str, Path, os.PathLike], optional
            Path to stored noise, if stored separately (in json format).
        initial_predictions_root : Union[str, Path, os.PathLike], optional
            Path to the dataset containing the initial predictions. If provided, the initial predictions will be used
            as the input of the reconstruction network. Default is ``None``.
        dataset_format : str, optional
            The format of the dataset. For example, ``'custom_dataset'`` or ``'public_dataset_name'``.
            Default is ``None``.
        sample_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the slices should be loaded. When creating
            subsampled datasets either set sample_rates (sample by slices) or volume_sample_rates (sample by volumes)
            but not both.
        volume_sample_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the volumes should be loaded. When creating
            subsampled datasets either set sample_rates (sample by slices) or volume_sample_rates (sample by volumes)
            but not both.
        use_dataset_cache : bool, optional
            Whether to cache dataset metadata. This is very useful for large datasets.
        dataset_cache_file : Union[str, Path, os.PathLike], optional
            A file in which to cache dataset information for faster load times.
        num_cols : Optional[Tuple[int]], optional
            If provided, only slices with the desired number of columns will be considered.
        consecutive_slices : int, optional
            An int (>0) that determine the amount of consecutive slices of the file to be loaded at the same time.
            Default is ``1``, loading single slices.
        data_saved_per_slice : bool, optional
            Whether the data is saved per slice or per volume.
        n2r_supervised_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the subjects should be loaded for Noise to
            Reconstruction (N2R) supervised loss, if N2R is enabled. Default is ``0.0``.
        complex_target : bool, optional
            Whether to use a complex target or not. Default is ``False``.
        log_images_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the subjects should be logged as images. Default is
            ``1.0``.
        transform : Optional[Callable], optional
            A sequence of callable objects that preprocesses the raw data into appropriate form. The transform function
            should take ``kspace``, ``coil sensitivity maps``, ``quantitative maps``, ``mask``, ``initial prediction``,
            ``target``, ``attributes``, ``filename``, and ``slice number`` as inputs. ``target`` may be null for test
            data. Default is ``None``.
        sequence : str, optional
            Sequence of the dataset. For example, ``MEGRE`` or ``FUTURE_SEQUENCES``.
        segmentation_mask_root : Union[str, Path, os.PathLike], optional
            Path to stored segmentation masks, if stored separately.
        kspace_scaling_factor : float, optional
            A float that scales the kspace. Default is ``1.0``.
        """
        super().__init__(
            root,
            coil_sensitivity_maps_root,
            mask_root,
            noise_root,
            initial_predictions_root,
            dataset_format,
            sample_rate,
            volume_sample_rate,
            use_dataset_cache,
            dataset_cache_file,
            num_cols,
            consecutive_slices,
            data_saved_per_slice,
            n2r_supervised_rate,
            complex_target,
            log_images_rate,
            transform,
            **kwargs,
        )
        if sequence not in ("MEGRE", "FUTURE_SEQUENCES"):
            warnings.warn(
                'Sequence should be either "MEGRE" or "FUTURE_SEQUENCES". '
                f'Found {sequence}. If you are using this dataset for reconstruction, ignore this warning.'
                'If you are using this dataset for quantitative mapping, please use the correct sequence.'
            )
        self.sequence = sequence
        self.segmentation_mask_root = segmentation_mask_root
        self.kspace_scaling_factor = kspace_scaling_factor

    def __getitem__(self, i: int):  # noqa: MC0001
        """Get item from :class:`qMRIDataset`."""
        raise NotImplementedError


class AHEADqMRIDataset(qMRIDataset):
    """Supports the AHEAD dataset for quantitative MRI.

    .. note::
        Extends :class:`atommic.collections.quantitative.data.qMRIDataset`.
    """

    def __getitem__(self, i: int):  # noqa: MC0001
        """Get item from :class:`AHEADqMRIDataset`."""
        fname, dataslice, metadata = self.examples[i]
        with h5py.File(fname, "r") as hf:
            kspace = self.get_consecutive_slices(hf, "kspace", dataslice).astype(np.complex64)

            kspace = kspace / self.kspace_scaling_factor

            if "sensitivity_map" in hf:
                sensitivity_map = self.get_consecutive_slices(hf, "sensitivity_map", dataslice).astype(np.complex64)
            elif not is_none(self.coil_sensitivity_maps_root):
                coil_sensitivity_maps_root = self.coil_sensitivity_maps_root
                split_dir = str(fname).split("/")
                # check if exists
                if not os.path.exists(Path(f"{coil_sensitivity_maps_root}/{split_dir[-2]}/{fname.name}")):
                    # find to what depth the coil_sensitivity_maps_root directory is nested
                    for j in range(len(split_dir)):
                        # get the coil_sensitivity_maps_root directory name
                        coil_sensitivity_maps_root = Path(f"{self.coil_sensitivity_maps_root}/{split_dir[-j]}/")
                        if os.path.exists(coil_sensitivity_maps_root / Path(split_dir[-2]) / fname.name):
                            break

                with h5py.File(
                    Path(coil_sensitivity_maps_root) / Path(split_dir[-2]) / fname.name, "r"  # type: ignore
                ) as sf:
                    if "sensitivity_map" in sf or "sensitivity_map" in next(iter(sf.keys())):
                        sensitivity_map = (
                            self.get_consecutive_slices(sf, "sensitivity_map", dataslice)
                            .squeeze()
                            .astype(np.complex64)
                        )
            else:
                sensitivity_map = np.array([])

            if "mask" in hf:
                mask = np.asarray(self.get_consecutive_slices(hf, "mask", dataslice))
                if mask.ndim == 3:
                    mask = mask[dataslice]
            elif not is_none(self.mask_root):
                with h5py.File(Path(self.mask_root) / fname.name, "r") as mf:  # type: ignore
                    mask = np.asarray(self.get_consecutive_slices(mf, "mask", dataslice))
            else:
                mask = np.empty([])

            if "anatomy_mask" in hf:
                anatomy_mask = np.asarray(self.get_consecutive_slices(hf, "anatomy_mask", dataslice))
                if anatomy_mask.ndim == 3:
                    anatomy_mask = anatomy_mask[dataslice]
            elif not is_none(self.segmentation_mask_root):
                with h5py.File(Path(self.segmentation_mask_root) / fname.name, "r") as mf:  # type: ignore
                    anatomy_mask = np.asarray(self.get_consecutive_slices(mf, "anatomy_mask", dataslice))
            else:
                anatomy_mask = np.empty([])

            mask = [mask, anatomy_mask]

            prediction = np.empty([])
            if not is_none(self.initial_predictions_root):
                with h5py.File(Path(self.initial_predictions_root) / fname.name, "r") as ipf:  # type: ignore
                    rkey = "reconstruction" if "reconstruction" in ipf else "initial_prediction"
                    prediction = self.get_consecutive_slices(ipf, rkey, dataslice).squeeze().astype(np.complex64)
            elif "reconstruction" in hf or "initial_prediction" in hf:
                rkey = "reconstruction" if "reconstruction" in hf else "initial_prediction"
                prediction = self.get_consecutive_slices(hf, rkey, dataslice).squeeze().astype(np.complex64)

            # find key containing "reconstruction_"
            rkey = re.findall(r"reconstruction_(.*)", str(hf.keys()))  # type: ignore
            self.recons_key = "reconstruction_" + rkey[0] if rkey else "target"
            if "reconstruction_rss" in self.recons_key:
                self.recons_key = "reconstruction_rss"
            elif "reconstruction_sense" in hf:
                self.recons_key = "reconstruction_sense"

            if self.complex_target:
                target = None
            else:
                # find key containing "reconstruction_"
                rkey = re.findall(r"reconstruction_(.*)", str(hf.keys()))  # type: ignore
                self.recons_key = "reconstruction_" + rkey[0] if rkey else "target"
                if "reconstruction_rss" in self.recons_key:
                    self.recons_key = "reconstruction_rss"
                elif "reconstruction_sense" in hf:
                    self.recons_key = "reconstruction_sense"
                target = self.get_consecutive_slices(hf, self.recons_key, dataslice) if self.recons_key in hf else None

            attrs = dict(hf.attrs)

            # get noise level for current slice, if metadata["noise_levels"] is not empty
            if "noise_levels" in metadata and len(metadata["noise_levels"]) > 0:
                metadata["noise"] = metadata["noise_levels"][dataslice]
            else:
                metadata["noise"] = 1.0

            attrs.update(metadata)

        if self.data_saved_per_slice:
            # arbitrary slice number for logging purposes
            dataslice = str(fname.name)  # type: ignore
            if "h5" in dataslice:  # type: ignore
                dataslice = dataslice.split(".h5")[0]  # type: ignore
            dataslice = int(dataslice.split("_")[-1])  # type: ignore

        attrs["log_image"] = bool(dataslice in self.indices_to_log) if not self.data_saved_per_slice else True

        if not is_none(self.sequence):
            return (
                (
                    kspace,
                    sensitivity_map,
                    mask,
                    prediction,
                    target,
                    attrs,
                    fname.name,
                    dataslice,
                )
                if self.transform is None
                else self.transform(
                    kspace,
                    sensitivity_map,
                    mask,
                    prediction,
                    target,
                    attrs,
                    fname.name,
                    dataslice,
                )
            )
        return (
            (
                kspace,
                sensitivity_map,
                np.empty([]),
                prediction,
                target,
                attrs,
                fname.name,
                dataslice,
            )
            if self.transform is None
            else self.transform(
                kspace,
                sensitivity_map,
                np.empty([]),
                prediction,
                target,
                attrs,
                fname.name,
                dataslice,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import os
import warnings
from pathlib import Path
from typing import Callable, Optional, Tuple, Union

import h5py
import nibabel as nib
import numpy as np

from atommic.collections.common.data.mri_loader import MRIDataset
from atommic.collections.common.parts.utils import is_none


class RSMRIDataset(MRIDataset):
    """A dataset class for accelerated-MRI reconstruction and MRI segmentation.

    Examples
    --------
    >>> from atommic.collections.multitask.rs.data.mrirs_loader import RSMRIDataset
    >>> dataset = RSMRIDataset(root='data/train', sample_rate=0.1)
    >>> print(len(dataset))
    100
    >>> kspace, imspace, coil_sensitivities, mask, initial_prediction, segmentation_labels, attrs, filename, \
    slice_num = dataset[0]
    >>> print(kspace.shape)
    np.array([30, 640, 368])

    .. note::
        Extends :class:`atommic.collections.common.data.MRIDataset`.
    """

    def __init__(
        self,
        root: Union[str, Path, os.PathLike],
        coil_sensitivity_maps_root: Union[str, Path, os.PathLike] = None,
        mask_root: Union[str, Path, os.PathLike] = None,
        noise_root: Union[str, Path, os.PathLike] = None,
        initial_predictions_root: Union[str, Path, os.PathLike] = None,
        dataset_format: str = None,
        sample_rate: Optional[float] = None,
        volume_sample_rate: Optional[float] = None,
        use_dataset_cache: bool = False,
        dataset_cache_file: Union[str, Path, os.PathLike] = None,
        num_cols: Optional[Tuple[int]] = None,
        consecutive_slices: int = 1,
        data_saved_per_slice: bool = False,
        n2r_supervised_rate: Optional[float] = 0.0,
        complex_target: bool = False,
        log_images_rate: Optional[float] = 1.0,
        transform: Optional[Callable] = None,
        segmentations_root: Union[str, Path, os.PathLike] = None,
        segmentation_classes: int = 2,
        segmentation_classes_to_remove: Optional[Tuple[int]] = None,
        segmentation_classes_to_combine: Optional[Tuple[int]] = None,
        segmentation_classes_to_separate: Optional[Tuple[int]] = None,
        segmentation_classes_thresholds: Optional[Tuple[float]] = None,
        complex_data: bool = True,
        **kwargs,
    ):
        """Inits :class:`RSMRIDataset`.

        Parameters
        ----------
        root : Union[str, Path, os.PathLike]
            Path to the dataset.
        sense_root : Union[str, Path, os.PathLike], optional
            Path to the coil sensitivities maps dataset, if stored separately.
        mask_root : Union[str, Path, os.PathLike], optional
            Path to stored masks, if stored separately.
        noise_root : Union[str, Path, os.PathLike], optional
            Path to stored noise, if stored separately (in json format).
        initial_predictions_root : Union[str, Path, os.PathLike], optional
            Path to the dataset containing the initial predictions. If provided, the initial predictions will be used
            as the input of the reconstruction network. Default is ``None``.
        dataset_format : str, optional
            The format of the dataset. For example, ``'custom_dataset'`` or ``'public_dataset_name'``.
            Default is ``None``.
        sample_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the slices should be loaded. When creating
            subsampled datasets either set sample_rates (sample by slices) or volume_sample_rates (sample by volumes)
            but not both.
        volume_sample_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the volumes should be loaded. When creating
            subsampled datasets either set sample_rates (sample by slices) or volume_sample_rates (sample by volumes)
            but not both.
        use_dataset_cache : bool, optional
            Whether to cache dataset metadata. This is very useful for large datasets.
        dataset_cache_file : Union[str, Path, os.PathLike], optional
            A file in which to cache dataset information for faster load times.
        num_cols : Optional[Tuple[int]], optional
            If provided, only slices with the desired number of columns will be considered.
        consecutive_slices : int, optional
            An int (>0) that determine the amount of consecutive slices of the file to be loaded at the same time.
            Default is ``1``, loading single slices.
        data_saved_per_slice : bool, optional
            Whether the data is saved per slice or per volume.
        n2r_supervised_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the subjects should be loaded for Noise to
            Reconstruction (N2R) supervised loss, if N2R is enabled. Default is ``0.0``.
        complex_target : bool, optional
            Whether to use a complex target or not. Default is ``False``.
        log_images_rate : Optional[float], optional
            A float between 0 and 1. This controls what fraction of the subjects should be logged as images. Default is
            ``1.0``.
        transform : Optional[Callable], optional
            A sequence of callable objects that preprocesses the raw data into appropriate form. The transform function
            should take ``kspace``, ``coil sensitivity maps``, ``mask``, ``initial prediction``, ``segmentation``,
            ``target``, ``attributes``, ``filename``, and ``slice number`` as inputs. ``target`` may be null for test
            data. Default is ``None``.
        segmentations_root : Union[str, Path, os.PathLike], optional
            Path to the dataset containing the segmentations.
        segmentation_classes : int, optional
            The number of segmentation classes. Default is ``2``.
        segmentation_classes_to_remove : Optional[Tuple[int]], optional
            A tuple of segmentation classes to remove. For example, if the dataset contains segmentation classes
            0, 1, 2, 3, and 4, and you want to remove classes 1 and 3, set this to ``(1, 3)``. Default is ``None``.
        segmentation_classes_to_combine : Optional[Tuple[int]], optional
            A tuple of segmentation classes to combine. For example, if the dataset contains segmentation classes
            0, 1, 2, 3, and 4, and you want to combine classes 1 and 3, set this to ``(1, 3)``. Default is ``None``.
        segmentation_classes_to_separate : Optional[Tuple[int]], optional
            A tuple of segmentation classes to separate. For example, if the dataset contains segmentation classes
            0, 1, 2, 3, and 4, and you want to separate class 1 into 2 classes, set this to ``(1, 2)``.
            Default is ``None``.
        segmentation_classes_thresholds : Optional[Tuple[float]], optional
            A tuple of thresholds for the segmentation classes. For example, if the dataset contains segmentation
            classes 0, 1, 2, 3, and 4, and you want to set the threshold for class 1 to 0.5, set this to
            ``(0.5, 0.5, 0.5, 0.5, 0.5)``. Default is ``None``.
        complex_data : bool, optional
            Whether the data is complex. If ``False``, the data is assumed to be magnitude only. Default is ``True``.
        **kwargs : dict
            Additional keyword arguments.
        """
        super().__init__(
            root,
            coil_sensitivity_maps_root,
            mask_root,
            noise_root,
            initial_predictions_root,
            dataset_format,
            sample_rate,
            volume_sample_rate,
            use_dataset_cache,
            dataset_cache_file,
            num_cols,
            consecutive_slices,
            data_saved_per_slice,
            n2r_supervised_rate,
            complex_target,
            log_images_rate,
            transform,
            **kwargs,
        )

        self.segmentations_root = segmentations_root

        # Create random number generator used for consecutive slice selection and set consecutive slice amount
        self.consecutive_slices = consecutive_slices
        self.segmentation_classes = segmentation_classes
        self.segmentation_classes_to_remove = segmentation_classes_to_remove
        self.segmentation_classes_to_combine = segmentation_classes_to_combine
        self.segmentation_classes_to_separate = segmentation_classes_to_separate
        self.segmentation_classes_thresholds = segmentation_classes_thresholds
        self.complex_data = complex_data

    def process_segmentation_labels(self, segmentation_labels: np.ndarray) -> np.ndarray:  # noqa: MC0001
        """Processes segmentation labels to remove, combine, and separate classes.

        Parameters
        ----------
        segmentation_labels : np.ndarray
            The segmentation labels. The shape should be (num_slices, height, width) or (height, width).

        Returns
        -------
        np.ndarray
            The processed segmentation labels.
        """
        # find the dimension with the segmentation classes
        segmentation_labels_dim = segmentation_labels.ndim - 1
        for dim in range(segmentation_labels.ndim):
            if segmentation_labels.shape[dim] == self.segmentation_classes:
                segmentation_labels_dim = dim

        # move it to the last dimension
        segmentation_labels = np.moveaxis(segmentation_labels, segmentation_labels_dim, -1)

        # if we have a single slice, add a new dimension
        if segmentation_labels.ndim == 2:
            segmentation_labels = np.expand_dims(segmentation_labels, axis=0)

        # check if we need to remove any classes, e.g. background
        if self.segmentation_classes_to_remove is not None:
            segmentation_labels = np.delete(segmentation_labels, self.segmentation_classes_to_remove, axis=-1)

        # check if we need to combine any classes, e.g. White Matter and Gray Matter
        if self.segmentation_classes_to_combine is not None:
            segmentation_labels_to_combine = np.sum(
                segmentation_labels[..., self.segmentation_classes_to_combine], axis=-1, keepdims=True
            )
            segmentation_labels_to_keep = np.delete(segmentation_labels, self.segmentation_classes_to_combine, axis=-1)

            if self.segmentation_classes_to_remove is not None and 0 in self.segmentation_classes_to_remove:
                # if background is removed, we can stack the combined labels with the rest straight away
                segmentation_labels = np.concatenate(
                    [segmentation_labels_to_combine, segmentation_labels_to_keep], axis=-1
                )
            else:
                # if background is not removed, we need to add it back as new background channel
                segmentation_labels = np.concatenate(
                    [segmentation_labels[..., 0:1], segmentation_labels_to_combine, segmentation_labels_to_keep],
                    axis=-1,
                )

        # check if we need to separate any classes, e.g. pathologies from White Matter and Gray Matter
        if self.segmentation_classes_to_separate is not None:
            for x in self.segmentation_classes_to_separate:
                segmentation_class_to_separate = segmentation_labels[..., x]
                for i in range(segmentation_labels.shape[-1]):
                    if i == x:
                        continue
                    segmentation_labels[..., i][segmentation_class_to_separate > 0] = 0

        # threshold probability maps if any threshold is given
        if self.segmentation_classes_thresholds is not None:
            for i, voxel_thres in enumerate(self.segmentation_classes_thresholds):
                if voxel_thres is not None:
                    segmentation_labels[..., i][segmentation_labels[..., i] < voxel_thres] = 0
                    segmentation_labels[..., i][segmentation_labels[..., i] >= voxel_thres] = 1

        if self.consecutive_slices == 1:
            # bring the segmentation classes dimension back to the first dimension
            segmentation_labels = np.moveaxis(segmentation_labels, -1, 0)
        elif self.consecutive_slices > 1:
            # bring the segmentation classes dimension back to the second dimension
            segmentation_labels = np.moveaxis(segmentation_labels, -1, 1)

        return segmentation_labels

    def __getitem__(self, i: int):  # noqa: MC0001
        """Get item from :class:`RSMRIDataset`."""
        fname, dataslice, metadata = self.examples[i]
        with h5py.File(fname, "r") as hf:
            if self.complex_data:
                kspace = self.get_consecutive_slices(hf, "kspace", dataslice).astype(np.complex64)

                sensitivity_map = np.array([])
                if "sensitivity_map" in hf:
                    sensitivity_map = self.get_consecutive_slices(hf, "sensitivity_map", dataslice).astype(
                        np.complex64
                    )
                elif "maps" in hf:
                    sensitivity_map = self.get_consecutive_slices(hf, "maps", dataslice).astype(np.complex64)
                elif self.coil_sensitivity_maps_root is not None and self.coil_sensitivity_maps_root != "None":
                    coil_sensitivity_maps_root = self.coil_sensitivity_maps_root
                    split_dir = str(fname).split("/")
                    # check if exists
                    if not os.path.exists(Path(f"{coil_sensitivity_maps_root}/{split_dir[-2]}/{fname.name}")):
                        # find to what depth the coil_sensitivity_maps_root directory is nested
                        for j in range(len(split_dir)):
                            # get the coil_sensitivity_maps_root directory name
                            coil_sensitivity_maps_root = Path(f"{self.coil_sensitivity_maps_root}/{split_dir[-j]}/")
                            if os.path.exists(coil_sensitivity_maps_root / Path(split_dir[-2]) / fname.name):
                                break
                    # load coil sensitivity maps
                    with h5py.File(Path(coil_sensitivity_maps_root) / Path(split_dir[-2]) / fname.name, "r") as sf:
                        if "sensitivity_map" in sf or "sensitivity_map" in next(iter(sf.keys())):
                            sensitivity_map = (
                                self.get_consecutive_slices(sf, "sensitivity_map", dataslice)
                                .squeeze()
                                .astype(np.complex64)
                            )

                mask = None
                if "mask" in hf:
                    mask = np.asarray(self.get_consecutive_slices(hf, "mask", dataslice))
                    if mask.ndim == 3:
                        mask = mask[dataslice]
                elif self.mask_root is not None and self.mask_root != "None":
                    with h5py.File(Path(self.mask_root) / fname.name, "r") as mf:
                        mask = np.asarray(self.get_consecutive_slices(mf, "mask", dataslice))

                imspace = np.empty([])

            elif not self.complex_data:
                if "reconstruction_rss" in hf:
                    imspace = self.get_consecutive_slices(hf, "reconstruction_rss", dataslice)
                elif "reconstruction_sense" in hf:
                    imspace = self.get_consecutive_slices(hf, "reconstruction_sense", dataslice)
                elif "reconstruction" in hf:
                    imspace = self.get_consecutive_slices(hf, "reconstruction", dataslice)
                elif "target" in hf:
                    imspace = self.get_consecutive_slices(hf, "target", dataslice)
                else:
                    raise ValueError(
                        "Complex data has not been selected but no reconstruction data found in file. "
                        "Only 'reconstruction' key is supported."
                    )
                kspace = np.empty([])
                sensitivity_map = np.array([])
                mask = np.empty([])

            segmentation_labels = np.empty([])
            if self.segmentations_root is not None and self.segmentations_root != "None":
                with h5py.File(Path(self.segmentations_root) / fname.name, "r") as sf:
                    segmentation_labels = np.asarray(self.get_consecutive_slices(sf, "segmentation", dataslice))
                    segmentation_labels = self.process_segmentation_labels(segmentation_labels)
            elif "segmentation" in hf:
                segmentation_labels = np.asarray(self.get_consecutive_slices(hf, "segmentation", dataslice))
                segmentation_labels = self.process_segmentation_labels(segmentation_labels)

            initial_prediction = np.empty([])
            if not is_none(self.initial_predictions_root):
                with h5py.File(Path(self.initial_predictions_root) / fname.name, "r") as ipf:  # type: ignore
                    if "reconstruction" in hf:
                        initial_prediction = (
                            self.get_consecutive_slices(ipf, "reconstruction", dataslice)
                            .squeeze()
                            .astype(np.complex64)
                        )
                    elif "initial_prediction" in hf:
                        initial_prediction = (
                            self.get_consecutive_slices(ipf, "initial_prediction", dataslice)
                            .squeeze()
                            .astype(np.complex64)
                        )
            else:
                if "reconstruction" in hf:
                    initial_prediction = (
                        self.get_consecutive_slices(hf, "reconstruction", dataslice).squeeze().astype(np.complex64)
                    )
                elif "initial_prediction" in hf:
                    initial_prediction = (
                        self.get_consecutive_slices(hf, "initial_prediction", dataslice).squeeze().astype(np.complex64)
                    )

            attrs = dict(hf.attrs)

            # get noise level for current slice, if metadata["noise_levels"] is not empty
            if "noise_levels" in metadata and len(metadata["noise_levels"]) > 0:
                metadata["noise"] = metadata["noise_levels"][dataslice]
            else:
                metadata["noise"] = 1.0

            attrs.update(metadata)

        if sensitivity_map.shape != kspace.shape and sensitivity_map.ndim > 1:
            if sensitivity_map.ndim == 3:
                sensitivity_map = np.transpose(sensitivity_map, (2, 0, 1))
            elif sensitivity_map.ndim == 4:
                sensitivity_map = np.transpose(sensitivity_map, (0, 3, 1, 2))
            else:
                raise ValueError(
                    f"Sensitivity map has invalid dimensions {sensitivity_map.shape} compared to kspace {kspace.shape}"
                )

        attrs["log_image"] = bool(dataslice in self.indices_to_log)

        return (
            (
                kspace,
                imspace,
                sensitivity_map,
                mask,
                initial_prediction,
                segmentation_labels,
                attrs,
                fname.name,
                dataslice,
            )
            if self.transform is None
            else self.transform(
                kspace,
                imspace,
                sensitivity_map,
                mask,
                initial_prediction,
                segmentation_labels,
                attrs,
                fname.name,
                dataslice,
            )
        )


class SKMTEARSMRIDataset(RSMRIDataset):
    """Supports the SKM-TEA dataset for multitask accelerated MRI reconstruction and MRI segmentation.

    .. note::
        Extends :class:`atommic.collections.multitask.rs.data.mrirs_loader.RSMRIDataset`.
    """

    def __getitem__(self, i: int):  # noqa: MC0001
        """Get item from :class:`SKMTEARSMRIDataset`."""
        if not is_none(self.dataset_format):
            dataset_format = self.dataset_format.lower()  # type: ignore
            masking = "default"
            if "custom_masking" in dataset_format:
                masking = "custom"
                dataset_format = dataset_format.replace("custom_masking", "").strip("_")
        else:
            dataset_format = None
            masking = "custom"

        fname, dataslice, metadata = self.examples[i]
        with h5py.File(fname, "r") as hf:
            kspace = self.get_consecutive_slices(hf, "kspace", dataslice).astype(np.complex64)

            if not is_none(dataset_format) and dataset_format == "skm-tea-echo1":
                kspace = kspace[:, :, 0, :]
            elif not is_none(dataset_format) and dataset_format == "skm-tea-echo2":
                kspace = kspace[:, :, 1, :]
            elif not is_none(dataset_format) and dataset_format == "skm-tea-echo1+echo2":
                kspace = kspace[:, :, 0, :] + kspace[:, :, 1, :]
            elif not is_none(dataset_format) and dataset_format == "skm-tea-echo1+echo2-mc":
                kspace = np.concatenate([kspace[:, :, 0, :], kspace[:, :, 1, :]], axis=-1)
            else:
                warnings.warn(
                    f"Dataset format {dataset_format} is either not supported or set to None. "
                    "Using by default only the first echo."
                )
                kspace = kspace[:, :, 0, :]

            kspace = kspace[48:-48, 40:-40]

            sensitivity_map = self.get_consecutive_slices(hf, "maps", dataslice).astype(np.complex64)
            sensitivity_map = sensitivity_map[..., 0]

            sensitivity_map = sensitivity_map[48:-48, 40:-40]

            if masking == "custom":
                mask = np.array([])
            else:
                masks = hf["masks"]
                mask = {}
                for key, val in masks.items():
                    mask[key.split("_")[-1].split(".")[0]] = np.asarray(val)

            # get the file format of the segmentation files
            segmentation_labels = nib.load(
                Path(self.segmentations_root) / Path(str(fname.name.split(".")[0]) + ".nii.gz")  # type: ignore
            ).get_fdata()

            # get a slice
            segmentation_labels = self.get_consecutive_slices({"seg": segmentation_labels}, "seg", dataslice)

            # Get the segmentation labels. They are valued as follows:
            # 0: Patellar Cartilage
            patellar_cartilage = np.zeros_like(segmentation_labels)
            patellar_cartilage[segmentation_labels == 1] = 1
            # 1: Femoral Cartilage
            femoral_cartilage = np.zeros_like(segmentation_labels)
            femoral_cartilage[segmentation_labels == 2] = 1
            # 2: Lateral Tibial Cartilage
            lateral_tibial_cartilage = np.zeros_like(segmentation_labels)
            lateral_tibial_cartilage[segmentation_labels == 3] = 1
            # 3: Medial Tibial Cartilage
            medial_tibial_cartilage = np.zeros_like(segmentation_labels)
            medial_tibial_cartilage[segmentation_labels == 4] = 1
            # 4: Lateral Meniscus
            lateral_meniscus = np.zeros_like(segmentation_labels)
            lateral_meniscus[segmentation_labels == 5] = 1
            # 5: Medial Meniscus
            medial_meniscus = np.zeros_like(segmentation_labels)
            medial_meniscus[segmentation_labels == 6] = 1
            # combine Lateral Tibial Cartilage and Medial Tibial Cartilage
            tibial_cartilage = lateral_tibial_cartilage + medial_tibial_cartilage
            # combine Lateral Meniscus and Medial Meniscus
            medial_meniscus = lateral_meniscus + medial_meniscus

            if self.consecutive_slices > 1:
                segmentation_labels_dim = 1
            else:
                segmentation_labels_dim = 0

            # stack the labels in the last dimension
            segmentation_labels = np.stack(
                [patellar_cartilage, femoral_cartilage, tibial_cartilage, medial_meniscus],
                axis=segmentation_labels_dim,
            )

            # TODO: This is hardcoded on the SKM-TEA side, how to generalize this?
            # We need to crop the segmentation labels in the frequency domain to reduce the FOV.
            segmentation_labels = np.fft.fftshift(np.fft.fft2(segmentation_labels))
            segmentation_labels = segmentation_labels[:, 48:-48, 40:-40]
            segmentation_labels = np.fft.ifft2(np.fft.ifftshift(segmentation_labels)).real

            imspace = np.empty([])

            initial_prediction = np.empty([])
            attrs = dict(hf.attrs)

            # get noise level for current slice, if metadata["noise_levels"] is not empty
            if "noise_levels" in metadata and len(metadata["noise_levels"]) > 0:
                metadata["noise"] = metadata["noise_levels"][dataslice]
            else:
                metadata["noise"] = 1.0

            attrs.update(metadata)

        kspace = np.transpose(kspace, (2, 0, 1))
        sensitivity_map = np.transpose(sensitivity_map.squeeze(), (2, 0, 1))

        attrs["log_image"] = bool(dataslice in self.indices_to_log)

        return (
            (
                kspace,
                imspace,
                sensitivity_map,
                mask,
                initial_prediction,
                segmentation_labels,
                attrs,
                fname.name,
                dataslice,
            )
            if self.transform is None
            else self.transform(
                kspace,
                imspace,
                sensitivity_map,
                mask,
                initial_prediction,
                segmentation_labels,
                attrs,
                fname.name,
                dataslice,
            )
        )



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.reconstruction.nn.ccnn import CascadeNet  # noqa: F401
from atommic.collections.reconstruction.nn.cirim import CIRIM  # noqa: F401
from atommic.collections.reconstruction.nn.crnn import CRNNet  # noqa: F401
from atommic.collections.reconstruction.nn.dunet import DUNet  # noqa: F401
from atommic.collections.reconstruction.nn.jointicnet import JointICNet  # noqa: F401
from atommic.collections.reconstruction.nn.kikinet import KIKINet  # noqa: F401
from atommic.collections.reconstruction.nn.lpd import LPDNet  # noqa: F401
from atommic.collections.reconstruction.nn.modl import MoDL  # noqa: F401
from atommic.collections.reconstruction.nn.multidomainnet import MultiDomainNet  # noqa: F401
from atommic.collections.reconstruction.nn.proximal_gradient import ProximalGradient  # noqa: F401
from atommic.collections.reconstruction.nn.recurrentvarnet import RecurrentVarNet  # noqa: F401
from atommic.collections.reconstruction.nn.unet import UNet  # noqa: F401
from atommic.collections.reconstruction.nn.varnet import VarNet  # noqa: F401
from atommic.collections.reconstruction.nn.vsnet import VSNet  # noqa: F401
from atommic.collections.reconstruction.nn.xpdnet import XPDNet  # noqa: F401
from atommic.collections.reconstruction.nn.zf import ZF  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch import Tensor

from atommic.collections.common.parts.fft import ifft2
from atommic.collections.common.parts.utils import coil_combination_method
from atommic.collections.quantitative.nn.base import BaseqMRIReconstructionModel
from atommic.core.classes.common import typecheck

__all__ = ["qZF"]


class qZF(BaseqMRIReconstructionModel):
    """Abstract class for returning the initial estimates of the quantitative maps."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        # init superclass
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)
        quantitative_module_dimensionality = cfg_dict.get("quantitative_module_dimensionality")
        if quantitative_module_dimensionality != 2:
            raise ValueError(
                f"Only 2D is currently supported for qMRI models.Found {quantitative_module_dimensionality}"
            )

        self.use_reconstruction_module = cfg_dict.get("use_reconstruction_module")

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        R2star_map_init: torch.Tensor,
        S0_map_init: torch.Tensor,
        B0_map_init: torch.Tensor,
        phi_map_init: torch.Tensor,
        TEs: List,  # pylint: disable=unused-argument
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        initial_prediction: torch.Tensor,
        anatomy_mask: torch.Tensor,  # pylint: disable=unused-argument
        sampling_mask: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> Union[List[List[Tensor]], List[Tensor]]:
        """
        Forward pass of the network.

        Parameters
        ----------
        R2star_map_init : torch.Tensor
            Initial R2* map of shape [batch_size, n_x, n_y].
        S0_map_init : torch.Tensor
            Initial S0 map of shape [batch_size, n_x, n_y].
        B0_map_init : torch.Tensor
            Initial B0 map of shape [batch_size, n_x, n_y].
        phi_map_init : torch.Tensor
            Initial phase map of shape [batch_size, n_x, n_y].
        TEs : List
            List of echo times.
        y : torch.Tensor
            Subsampled k-space data of shape [batch_size, n_echoes, n_coils, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
        initial_prediction : torch.Tensor
            Initial prediction of shape [batch_size, n_x, n_y, 2].
        anatomy_mask : torch.Tensor
            Brain mask of shape [batch_size, 1, n_x, n_y, 1].
        sampling_mask : torch.Tensor
            Sampling mask of shape [batch_size, 1, n_x, n_y, 1].
        sigma : float
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        List of list of torch.Tensor or torch.Tensor
             If self.accumulate_loss is True, returns a list of all intermediate predictions.
             If False, returns the final estimate.
        """
        if self.use_reconstruction_module:
            reconstruction_prediction = torch.stack(
                [
                    torch.view_as_complex(
                        coil_combination_method(
                            ifft2(
                                y[:, echo, ...].clone(), self.fft_centered, self.fft_normalization, self.spatial_dims
                            ),
                            sensitivity_maps,
                            method=self.coil_combination_method,
                            dim=self.coil_dim - 1,
                        )
                    )
                    for echo in range(y.shape[1])
                ],
                dim=1,
            )
        else:
            reconstruction_prediction = initial_prediction

        return [
            reconstruction_prediction,
            R2star_map_init,
            S0_map_init,
            B0_map_init,
            phi_map_init,
        ]



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Any, Dict, Tuple, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch import nn

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import coil_combination_method
from atommic.collections.multitask.rs.nn.base import BaseMRIReconstructionSegmentationModel
from atommic.collections.multitask.rs.nn.idslr_base.idslr_block import DC, UnetDecoder, UnetEncoder
from atommic.collections.reconstruction.nn.rim_base.conv_layers import ConvNonlinear
from atommic.core.classes.common import typecheck

__all__ = ["SegNet"]


class SegNet(BaseMRIReconstructionSegmentationModel):
    """Implementation of the Segmentation Network MRI, as described in, as presented in [Sun2019]_.

    References
    ----------
    .. [Sun2019] Sun, L., Fan, Z., Ding, X., Huang, Y., Paisley, J. (2019). Joint CS-MRI Reconstruction and
        Segmentation with a Unified Deep Network. In: Chung, A., Gee, J., Yushkevich, P., Bao, S. (eds) Information
        Processing in Medical Imaging. IPMI 2019. Lecture Notes in Computer Science(), vol 11492. Springer, Cham.
        https://doi.org/10.1007/978-3-030-20351-1_38

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`SegNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object.
        trainer : Trainer, optional
            PyTorch Lightning trainer object. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.use_reconstruction_module = cfg_dict.get("use_reconstruction_module", True)

        self.dimensionality = cfg_dict.get("dimensionality", 2)
        if self.dimensionality != 2:
            raise NotImplementedError(f"Currently only 2D is supported for segmentation, got {self.dimensionality}D.")

        self.input_channels = cfg_dict.get("input_channels", 2)
        reconstruction_out_chans = cfg_dict.get("reconstruction_module_output_channels", 2)
        segmentation_out_chans = cfg_dict.get("segmentation_module_output_channels", 1)
        chans = cfg_dict.get("channels", 32)
        num_pools = cfg_dict.get("num_pools", 4)
        drop_prob = cfg_dict.get("drop_prob", 0.0)
        normalize = cfg_dict.get("normalize", False)
        padding = cfg_dict.get("padding", False)
        padding_size = cfg_dict.get("padding_size", 11)
        self.norm_groups = cfg_dict.get("norm_groups", 2)
        num_cascades = cfg_dict.get("num_cascades", 5)

        self.reconstruction_encoder = nn.ModuleList(
            [
                UnetEncoder(
                    chans=chans,
                    num_pools=num_pools,
                    in_chans=self.input_channels,
                    drop_prob=drop_prob,
                    normalize=normalize,
                    padding=padding,
                    padding_size=padding_size,
                    norm_groups=self.norm_groups,
                )
                for _ in range(num_cascades)
            ]
        )
        self.reconstruction_decoder = nn.ModuleList(
            [
                UnetDecoder(
                    chans=chans,
                    num_pools=num_pools,
                    out_chans=reconstruction_out_chans,
                    drop_prob=drop_prob,
                    normalize=normalize,
                    padding=padding,
                    padding_size=padding_size,
                    norm_groups=self.norm_groups,
                )
                for _ in range(num_cascades)
            ]
        )
        self.segmentation_decoder = nn.ModuleList(
            [
                UnetDecoder(
                    chans=chans,
                    num_pools=num_pools,
                    out_chans=segmentation_out_chans,
                    drop_prob=drop_prob,
                    normalize=normalize,
                    padding=padding,
                    padding_size=padding_size,
                    norm_groups=self.norm_groups,
                )
                for _ in range(num_cascades)
            ]
        )

        self.segmentation_final_layer = torch.nn.Sequential(
            ConvNonlinear(
                segmentation_out_chans * num_cascades,
                segmentation_out_chans,
                conv_dim=cfg_dict.get("segmentation_final_layer_conv_dim", 2),
                kernel_size=cfg_dict.get("segmentation_final_layer_kernel_size", 3),
                dilation=cfg_dict.get("segmentation_final_layer_dilation", 1),
                bias=cfg_dict.get("segmentation_final_layer_bias", False),
                nonlinear=cfg_dict.get("segmentation_final_layer_nonlinear", "relu"),
            )
        )

        self.magnitude_input = cfg_dict.get("magnitude_input", True)
        self.normalize_segmentation_output = cfg_dict.get("normalize_segmentation_output", True)

        self.dc = DC()

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        init_reconstruction_pred: torch.Tensor,
        target_reconstruction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> Tuple[Any, Any]:
        """Forward pass of :class:`SegNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        init_reconstruction_pred : torch.Tensor
            Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2]
        target_reconstruction : torch.Tensor
            Target reconstruction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        Tuple[Union[List, torch.Tensor], torch.Tensor]
            Tuple containing the predicted reconstruction and segmentation.
        """
        if self.consecutive_slices > 1:
            batch, slices = y.shape[0], y.shape[1]
            y = y.reshape(y.shape[0] * y.shape[1], *y.shape[2:])
            sensitivity_maps = sensitivity_maps.reshape(
                sensitivity_maps.shape[0] * sensitivity_maps.shape[1],
                *sensitivity_maps.shape[2:],
            )
            mask = mask.reshape(mask.shape[0] * mask.shape[1], *mask.shape[2:])

        # In case of deviating number of coils, we need to pad up to maximum number of coils == number of input \
        # channels for the reconstruction module
        num_coils = y.shape[1]
        if num_coils * 2 != self.input_channels:
            num_coils_to_add = (self.input_channels - num_coils * 2) // 2
            dummy_coil_data = torch.zeros_like(torch.movedim(y, self.coil_dim, 0)[0]).unsqueeze(self.coil_dim)
            for _ in range(num_coils_to_add):
                y = torch.cat([y, dummy_coil_data], dim=self.coil_dim)
                sensitivity_maps = torch.cat([sensitivity_maps, dummy_coil_data], dim=self.coil_dim)

        y_prediction = y.clone()
        pred_segmentations = []
        for re, rd, sd in zip(self.reconstruction_encoder, self.reconstruction_decoder, self.segmentation_decoder):
            init_reconstruction_pred = ifft2(
                y_prediction, self.fft_centered, self.fft_normalization, self.spatial_dims
            )
            output = re(init_reconstruction_pred)
            reconstruction_encoder_prediction, padding_size = output[0].copy(), output[2]

            pred_segmentation_input = reconstruction_encoder_prediction
            if self.magnitude_input:
                pred_segmentation_input = [torch.abs(x) for x in pred_segmentation_input]

            pred_segmentations.append(sd(pred_segmentation_input, iscomplex=False, pad_sizes=padding_size))
            reconstruction_decoder_prediction = rd(*output)
            reconstruction_decoder_prediction_kspace = fft2(
                reconstruction_decoder_prediction,
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            y_prediction = self.dc(reconstruction_decoder_prediction_kspace, y, mask)

        pred_reconstruction = self.process_intermediate_pred(y_prediction, sensitivity_maps, True)

        pred_segmentation = self.segmentation_final_layer(torch.cat(pred_segmentations, dim=1))

        if self.normalize_segmentation_output:
            pred_segmentation = (pred_segmentation - pred_segmentation.min()) / (
                pred_segmentation.max() - pred_segmentation.min()
            )

        pred_segmentation = torch.abs(pred_segmentation)

        pred_segmentations.append(pred_segmentation)

        if self.consecutive_slices > 1:
            # get batch size and number of slices from y, because if the reconstruction module is used they will not
            # be saved before
            pred_reconstruction = pred_reconstruction.view([batch, slices, *pred_reconstruction.shape[1:]])
            pred_segmentations = [x.view([batch, slices, *x.shape[1:]]) for x in pred_segmentations]

        return pred_reconstruction, pred_segmentations

    def process_segmentation_loss(self, target: torch.Tensor, prediction: torch.Tensor, attrs: Dict) -> Dict:
        """Processes the segmentation loss.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, nr_classes, n_x, n_y].
        prediction : torch.Tensor
            Prediction of shape [batch_size, nr_classes, n_x, n_y].
        attrs : Dict
            Attributes of the data with pre normalization values.

        Returns
        -------
        Dict
            Dictionary containing the (multiple) loss values. For example, if the cross entropy loss and the dice loss
            are used, the dictionary will contain the keys ``cross_entropy_loss``, ``dice_loss``, and
            (combined) ``segmentation_loss``.
        """
        if self.unnormalize_loss_inputs:
            target, prediction = self.__unnormalize_for_loss_or_log__(  # type: ignore
                target, prediction, None, attrs, attrs["r"]
            )
        losses = {}
        for name, loss_func in self.segmentation_losses.items():
            cascades_loss = []
            for i in range(len(prediction)):  # pylint: disable=consider-using-enumerate
                loss = loss_func(target, prediction[i])
                if isinstance(loss, tuple):
                    # In case of the dice loss, the loss is a tuple of the form (dice, dice loss)
                    loss = loss[1]
                cascades_loss.append(loss)
            losses[name] = torch.stack(cascades_loss).mean().to(target.device)
        return self.total_segmentation_loss(**losses) * self.total_segmentation_loss_weight

    def process_intermediate_pred(
        self,
        prediction: Union[list, torch.Tensor],
        sensitivity_maps: torch.Tensor,
        do_coil_combination: bool = False,
    ) -> torch.Tensor:
        """Processes the intermediate prediction.

        Parameters
        ----------
        prediction : torch.Tensor
            Intermediate prediction. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        do_coil_combination : bool
            Whether to do coil combination. In this case the prediction is in k-space. Default is ``False``.

        Returns
        -------
        torch.Tensor, shape [batch_size, n_x, n_y, 2]
            Processed prediction.
        """
        # Take the last time step of the prediction
        if do_coil_combination:
            prediction = ifft2(
                prediction,
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            )
            prediction = coil_combination_method(
                prediction, sensitivity_maps, method=self.coil_combination_method, dim=self.coil_dim
            )
        prediction = torch.view_as_complex(prediction)
        return prediction



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/collections/common/losses/aggregator.py

from typing import List

import torch

from atommic.core.classes.common import typecheck
from atommic.core.classes.loss import Loss
from atommic.core.neural_types.elements import LossType
from atommic.core.neural_types.neural_type import NeuralType

__all__ = ["AggregatorLoss"]


class AggregatorLoss(Loss):
    """Aggregates multiple losses into a single loss.

    Examples
    --------
    >>> from atommic.collections.common.losses.aggregator import AggregatorLoss
    >>> loss = AggregatorLoss(num_inputs=2)
    >>> loss(loss_1=torch.tensor(1.0), loss_2=torch.tensor(2.0))
    tensor(3.)
    """

    @property
    def input_types(self):
        """Returns definitions of module input ports."""
        return {f"loss_{str(i + 1)}": NeuralType(elements_type=LossType()) for i in range(self._num_losses)}

    @property
    def output_types(self):
        """Returns definitions of module output ports."""
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, num_inputs: int = 2, weights: List[float] = None):
        """Inits :class:`AggregatorLoss`.

        Parameters
        ----------
        num_inputs : int
            Number of losses to be summed.
        weights : List[float]
            Weights to be applied to each loss. If None, all losses are weighted equally.
        """
        super().__init__()
        self._num_losses = num_inputs
        if weights is not None and len(weights) != num_inputs:
            raise ValueError("Length of weights should be equal to the number of inputs (num_inputs)")
        self._weights = weights

    @typecheck()
    def forward(self, **kwargs):
        """Computes the sum of the losses."""
        values = [kwargs[x] for x in sorted(kwargs.keys())]
        loss = torch.zeros_like(values[0])
        for loss_idx, loss_value in enumerate(values):
            if self._weights is not None:
                loss = loss.add(loss_value, alpha=self._weights[loss_idx])
            else:
                loss = loss.add(loss_value)
        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/decorators/experimental.py

import wrapt

from atommic.utils import logging

__all__ = ["experimental"]


@wrapt.decorator
def experimental(wrapped, instance, args, kwargs):  # pylint: disable=unused-argument
    """Decorator to mark a class as experimental.

    Parameters
    ----------
    wrapped : function
        The function to be decorated.
    instance : object
        The instance of the class to be decorated.
    args : tuple
        The arguments passed to the function to be decorated.
    kwargs : dict
        The keyword arguments passed to the function to be decorated.
    """
    logging.warning(f"`{wrapped}` is experimental and not ready for production yet. Use at your own risk.")
    return wrapped(*args, **kwargs)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.common.losses.aggregator import AggregatorLoss  # noqa: F401
from atommic.collections.common.losses.wasserstein import SinkhornDistance  # noqa: F401

VALID_RECONSTRUCTION_LOSSES = ["l1", "mse", "ssim", "noise_aware", "wasserstein"]
VALID_SEGMENTATION_LOSSES = ["cross_entropy", "dice"]



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/Project-MONAI/MONAI/blob/dev/monai/losses/dice.py

import warnings
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from atommic.collections.common.parts.utils import is_none
from atommic.collections.segmentation.losses.utils import do_metric_reduction
from atommic.core.classes.loss import Loss


class Dice(Loss):
    """Wrapper for :py:class:`monai.losses.DiceLoss` to support multi-class and multi-label tasks.

    Compute average Dice loss between two tensors. It can support both multi-classes and multi-labels tasks.
    The data `input` (BNHW[D] where N is number of classes) is compared with ground truth `target` (BNHW[D]).

    Note that axis N of `input` is expected to be logits or probabilities for each class, if passing logits as input,
    must set `sigmoid=True` or `softmax=True`, or specifying `other_act`. And the same axis of `target`
    can be 1 or N (one-hot format).

    The `smooth_nr` and `smooth_dr` parameters are values added to the intersection and union components of
    the inter-over-union calculation to smooth results respectively, these values should be small.

    The original paper: Milletari, F. et. al. (2016) V-Net: Fully Convolutional Neural Networks forVolumetric
    Medical Image Segmentation, 3DV, 2016.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.segmentation.losses.dice import Dice
    >>> pred = torch.tensor([[[[0.1, 0.2, 0.3, 0.4, 0.5],
    ...                        [0.1, 0.2, 0.3, 0.4, 0.5],
    ...                        [0.1, 0.2, 0.3, 0.4, 0.5],
    ...                        [0.1, 0.2, 0.3, 0.4, 0.5],
    ...                        [0.1, 0.2, 0.3, 0.4, 0.5]]],
    ...                       [[[0.1, 0.2, 0.3, 0.4, 0.5],
    ...                         [0.1, 0.2, 0.3, 0.4, 0.5],
    ...                         [0.1, 0.2, 0.3, 0.4, 0.5],
    ...                         [0.1, 0.2, 0.3, 0.4, 0.5],
    ...                         [0.1, 0.2, 0.3, 0.4, 0.5]]]])
    >>> target = torch.tensor([[[[0, 0, 0, 0, 0],
    ...                          [0, 0, 0, 0, 0],
    ...                          [0, 0, 0, 0, 0],
    ...                          [0, 0, 0, 0, 0],
    ...                          [0, 0, 0, 0, 0]]],
    ...                         [[[1, 1, 1, 1, 1],
    ...                           [1, 1, 1, 1, 1],
    ...                           [1, 1, 1, 1, 1],
    ...                           [1, 1, 1, 1, 1],
    ...                           [1, 1, 1, 1, 1]]]])
    >>> dice = Dice(include_background=False, to_onehot_y=True, sigmoid=False, softmax=False)
    >>> dice(pred, target)
    tensor(0.5000)
    """

    def __init__(
        self,
        include_background: bool = True,
        to_onehot_y: bool = False,
        sigmoid: bool = True,
        softmax: bool = False,
        other_act: Optional[Callable] = None,
        squared_pred: bool = False,
        jaccard: bool = False,
        flatten: bool = False,
        reduction: str = "mean",
        smooth_nr: float = 1e-5,
        smooth_dr: float = 1e-5,
        batch: bool = True,
    ):
        """Inits :class:`Dice`.

        Parameters
        ----------
        include_background : bool
            whether to skip Dice computation on the first channel of the predicted output. Default is ``True``.
        to_onehot_y : bool
            Whether to convert `y` into the one-hot format. Default is ``False``.
        sigmoid : bool
            Whether to add sigmoid function to the input data. Default is ``True``.
        softmax : bool
            Whether to add softmax function to the input data. Default is ``False``.
        other_act : Callable
            Use this parameter if you want to apply another type of activation layer. Default is ``None``.
        squared_pred : bool
            Whether to square the prediction before calculating Dice. Default is ``False``.
        jaccard : bool
            Whether to compute Jaccard Index as a loss. Default is ``False``.
        flatten : bool
            Whether to flatten input data. Default is ``False``.
        reduction : str
            Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.
            'none': no reduction will be applied.
            'mean': the sum of the output will be divided by the number of elements in the output.
            'sum': the output will be summed.
            Default is ``mean``.
        smooth_nr : float
            A small constant added to the numerator to avoid `nan` when all items are 0. Default is ``1e-5``.
        smooth_dr : float
            A small constant added to the denominator to avoid `nan` when all items are 0. Default is ``1e-5``.
        batch : bool
            If True, compute Dice loss for each batch and return a tensor with shape (batch_size,).
            If False, compute Dice loss for the whole batch and return a tensor with shape (1,).
            Default is ``True``.
        """
        super().__init__()
        other_act = None if is_none(other_act) else other_act
        if other_act is not None and not callable(other_act):
            raise TypeError(f"other_act must be None or callable but is {type(other_act).__name__}.")
        if int(sigmoid) + int(softmax) + int(other_act is not None) > 1:
            raise ValueError(
                "Incompatible values: more than 1 of [sigmoid=True, softmax=True, other_act is not None]."
            )
        self.include_background = include_background
        self.to_onehot_y = to_onehot_y
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.other_act = other_act
        self.squared_pred = squared_pred
        self.jaccard = jaccard
        self.flatten = flatten
        self.reduction = reduction
        self.smooth_nr = float(smooth_nr)
        self.smooth_dr = float(smooth_dr)
        self.batch = batch

    def forward(self, target: torch.Tensor, _input: torch.Tensor) -> Tuple[Union[Tensor, Any], Tensor]:  # noqa: MC0001
        """Forward pass of :class:`Dice`.

        Parameters
        ----------
        _input: torch.Tensor
            Prediction of shape [BNHW[D]].
        target: torch.Tensor
            Ground truth of shape [BNHW[D]].

        Returns
        -------
        torch.Tensor
            Dice loss.
        """
        if isinstance(_input, np.ndarray):
            _input = torch.from_numpy(_input)
        if isinstance(target, np.ndarray):
            target = torch.from_numpy(target)

        if self.flatten:
            if target.dim() == 4:
                segmentation_classes_dim = 1
            else:
                segmentation_classes_dim = 0
            target = target.reshape(target.shape[segmentation_classes_dim], 1, -1)
            _input = _input.reshape(_input.shape[segmentation_classes_dim], 1, -1)

        if self.sigmoid:
            _input = torch.sigmoid(_input.float())

        n_pred_ch = _input.shape[1]
        if self.softmax:
            if n_pred_ch == 1:
                warnings.warn("single channel prediction, `softmax=True` ignored.")
            else:
                _input = torch.softmax(_input.float(), 1).to(_input)

        if self.other_act is not None:
            _input = self.other_act(_input)

        if self.to_onehot_y:
            if n_pred_ch == 1:
                warnings.warn("single channel prediction, `to_onehot_y=True` ignored.")
            else:
                target = one_hot(target, num_classes=n_pred_ch)

        if not self.include_background:
            if n_pred_ch == 1:
                warnings.warn("single channel prediction, `include_background=False` ignored.")
            else:
                # if skipping background, removing first channel
                target = target[:, 1:]
                _input = _input[:, 1:]

        if target.shape != _input.shape:
            raise AssertionError(f"ground truth has different shape ({target.shape}) from _input ({_input.shape})")

        # reducing only spatial dimensions (not batch nor channels)
        reduce_axis: List[int] = torch.arange(2, len(_input.shape)).tolist()
        if self.batch:
            # reducing spatial dimensions and batch
            reduce_axis = [0] + reduce_axis
        intersection = torch.sum(target * _input, dim=reduce_axis)
        if self.squared_pred:
            target = torch.pow(target, 2)
            _input = torch.pow(_input, 2)
        ground_o = torch.sum(target, dim=reduce_axis)
        pred_o = torch.sum(_input, dim=reduce_axis)
        denominator = ground_o + pred_o
        if self.jaccard:
            denominator = 2.0 * (denominator - intersection)
        dice_score = (2.0 * intersection + self.smooth_nr) / (denominator + self.smooth_dr)
        dice_score = torch.where(denominator > 0, dice_score, torch.tensor(1.0).to(pred_o.device))
        dice_score, _ = do_metric_reduction(dice_score, reduction=self.reduction)
        f: torch.Tensor = 1.0 - dice_score
        return dice_score, f


def one_hot(labels: torch.Tensor, num_classes: int, dtype: torch.dtype = torch.float, dim: int = 1) -> torch.Tensor:
    """Convert labels to one-hot representation.

    Parameters
    ----------
    labels: torch.Tensor
        the labels of shape [BNHW[D]].
    num_classes: int
        number of classes.
    dtype: torch.dtype
        the data type of the returned tensor.
    dim: int
        the dimension to expand the one-hot tensor.

    Returns
    -------
    torch.Tensor
        The one-hot representation of the labels.

    Examples
    --------
    >>> labels = torch.tensor([[[[0, 1, 2]]]])
    >>> one_hot(labels, num_classes=3)
    tensor([[[[1., 0., 0.],
                [0., 1., 0.],
                [0., 0., 1.]]]])
    """
    # if `dim` is bigger, add singleton dim at the end
    if labels.ndim < dim + 1:
        shape = list(labels.shape) + [1] * (dim + 1 - len(labels.shape))
        labels = torch.reshape(labels, shape)
    sh = list(labels.shape)
    sh[dim] = num_classes
    o = torch.zeros(size=sh, dtype=dtype, device=labels.device)
    labels = o.scatter_(dim=dim, index=labels.long(), value=1)
    return labels



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/collections/common/callbacks/ema.py

import contextlib
import copy
import os
import threading
from typing import Any, Dict, Iterable, Optional

import pytorch_lightning as pl
import torch
from pytorch_lightning import Callback
from pytorch_lightning.utilities.exceptions import MisconfigurationException
from pytorch_lightning.utilities.rank_zero import rank_zero_info


class EMA(Callback):
    """Implements Exponential Moving Averaging (EMA).

    When training a model, this callback will maintain moving averages of the trained parameters.
    When evaluating, we use the moving averages copy of the trained parameters.
    When saving, we save an additional set of parameters with the prefix `ema`.

    .. note::
        Extends :class:`pytorch_lightning.callbacks.Callback`.

    Examples
    --------
    >>> from atommic.collections.common.callbacks.ema import EMA
    >>> ema = EMA(decay=0.9999, validate_original_weights=False, every_n_steps=1, cpu_offload=False)
    >>> trainer = Trainer(callbacks=[ema])
    >>> trainer.fit(model)
    >>> trainer.test(model)
    """

    def __init__(
        self,
        decay: float,
        validate_original_weights: bool = False,
        every_n_steps: int = 1,
        cpu_offload: bool = False,
    ):
        """Inits :class:`EMA`.

        Parameters
        ----------
        decay : float
            The exponential decay used when calculating the moving average. Has to be between 0-1.
        validate_original_weights : bool
            Validate the original weights, as apposed to the EMA weights. Default is ``False``.
        every_n_steps : int
            Apply EMA every N steps. Default is ``1``.
        cpu_offload : bool
            Offload weights to CPU. Default is ``False``.
        """
        super().__init__()
        if not 0 <= decay <= 1:
            raise MisconfigurationException("EMA decay value must be between 0 and 1")
        self.decay = decay
        self.validate_original_weights = validate_original_weights
        self.every_n_steps = every_n_steps
        self.cpu_offload = cpu_offload

    def on_fit_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Initialize the EMA weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        pl_module : pl.LightningModule
            PyTorch Lightning Module.
        """
        device = pl_module.device if not self.cpu_offload else torch.device("cpu")
        trainer.optimizers = [
            EMAOptimizer(
                optim,
                device=device,
                decay=self.decay,
                every_n_steps=self.every_n_steps,
                current_step=trainer.global_step,
            )
            for optim in trainer.optimizers
            if not isinstance(optim, EMAOptimizer)
        ]

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Swap the model weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        pl_module : pl.LightningModule
            PyTorch Lightning Module.
        """
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def on_validation_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Swap back the model weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        pl_module : pl.LightningModule
            PyTorch Lightning Module.
        """
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def on_test_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Swap the model weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        pl_module : pl.LightningModule
            PyTorch Lightning Module.
        """
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def on_test_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Swap back the model weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        pl_module : pl.LightningModule
            PyTorch Lightning Module.
        """
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def _should_validate_ema_weights(self, trainer: "pl.Trainer") -> bool:
        """Check if we should validate the EMA weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        """
        return not self.validate_original_weights and self._ema_initialized(trainer)

    def _ema_initialized(self, trainer: "pl.Trainer") -> bool:
        """Check if EMA has been initialized.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        """
        return any(isinstance(optimizer, EMAOptimizer) for optimizer in trainer.optimizers)

    def swap_model_weights(self, trainer: "pl.Trainer", saving_ema_model: bool = False):
        """Swaps the model weights with the EMA weights.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        saving_ema_model : bool, optional
            Whether we are saving the EMA model. Default is ``False``.
        """
        for optimizer in trainer.optimizers:
            assert isinstance(optimizer, EMAOptimizer)
            optimizer.switch_main_parameter_weights(saving_ema_model)

    @contextlib.contextmanager
    def save_ema_model(self, trainer: "pl.Trainer"):
        """Saves an EMA copy of the model + EMA optimizer states for resume.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        """
        self.swap_model_weights(trainer, saving_ema_model=True)
        try:
            yield
        finally:
            self.swap_model_weights(trainer, saving_ema_model=False)

    @contextlib.contextmanager
    def save_original_optimizer_state(self, trainer: "pl.Trainer"):
        """Saves the original optimizer states for resume.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        """
        for optimizer in trainer.optimizers:
            assert isinstance(optimizer, EMAOptimizer)
            optimizer.save_original_optimizer_state = True
        try:
            yield
        finally:
            for optimizer in trainer.optimizers:
                optimizer.save_original_optimizer_state = False

    def on_load_checkpoint(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", checkpoint: Dict[str, Any]
    ) -> None:
        """Restore EMA weights when loading a checkpoint.

        Parameters
        ----------
        trainer : pl.Trainer
            PyTorch Lightning Trainer.
        pl_module : pl.LightningModule
            PyTorch Lightning Module.
        checkpoint : Dict[str, Any]
            Checkpoint dictionary.
        """
        checkpoint_callback = trainer.checkpoint_callback

        # Replace connector._ckpt_path with below to avoid calling into lightning's protected API
        ckpt_path = trainer.ckpt_path

        if ckpt_path and checkpoint_callback is not None and "atommic" in type(checkpoint_callback).__name__:
            ext = checkpoint_callback.FILE_EXTENSION
            if ckpt_path.endswith(f"-EMA{ext}"):
                rank_zero_info(
                    "loading EMA based weights. The callback will treat the loaded EMA weights "
                    "as the main weights and create a new EMA copy when training."
                )
                return
            ema_path = ckpt_path.replace(ext, f"-EMA{ext}")
            if os.path.exists(ema_path):
                ema_state_dict = torch.load(ema_path, map_location=torch.device("cpu"))

                checkpoint["optimizer_states"] = ema_state_dict["optimizer_states"]
                rank_zero_info("EMA state has been restored.")
            else:
                raise MisconfigurationException(
                    "Unable to find the associated EMA weights when re-loading, training "
                    f"will start with new EMA weights. Expected them to be at: {ema_path}",
                )


@torch.no_grad()
def ema_update(ema_model_tuple, current_model_tuple, decay):
    """Update EMA parameters.

    Parameters
    ----------
    ema_model_tuple : tuple
        EMA model parameters.
    current_model_tuple : tuple
        Current model parameters.
    decay : float
        Decay factor.
    """
    torch._foreach_mul_(ema_model_tuple, decay)  # pylint: disable=protected-access
    torch._foreach_add_(ema_model_tuple, current_model_tuple, alpha=(1.0 - decay))  # pylint: disable=protected-access


def run_ema_update_cpu(ema_model_tuple, current_model_tuple, decay, pre_sync_stream=None):
    """Run EMA update on CPU.

    Parameters
    ----------
    ema_model_tuple : tuple
        EMA model parameters.
    current_model_tuple : tuple
        Current model parameters.
    decay : float
        Decay factor.
    pre_sync_stream : torch.cuda.Stream, optional
        CUDA stream. Default is ``None``.
    """
    if pre_sync_stream is not None:
        pre_sync_stream.synchronize()
    ema_update(ema_model_tuple, current_model_tuple, decay)


class EMAOptimizer(torch.optim.Optimizer):
    r"""EMAOptimizer is a wrapper for torch.optim.Optimizer that computes Exponential Moving Average of parameters
    registered in the optimizer.

    EMA parameters are automatically updated after every step of the optimizer with the following formula:
    $$ ema\_weight = ema\_weight + (1 - decay) * (training\_weight - ema\_weight) $$

    To access EMA parameters, use ``swap_ema_weights()`` context manager to perform a temporary in-place swap of
    regular parameters with EMA parameters.

    .. note::
        EMAOptimizer is not compatible with APEX AMP O2.
        Extends :class:`torch.optim.Optimizer`.

    Returns
    -------
    torch.optim.Optimizer
        EMAOptimizer instance.

    Examples
    --------
    >>> model = Model().to(device)
    >>> opt = EMAOptimizer(opt, device, 0.9999)
    >>> for epoch in range(n_epochs):
    >>>     training_loop(model, opt)
    >>>     regular_eval_accuracy = evaluate(model)
    >>>     with opt.swap_ema_weights():
    >>>         ema_eval_accuracy = evaluate(model)
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        decay: float = 0.9999,
        every_n_steps: int = 1,
        current_step: int = 0,
        stream: Optional[torch.cuda.Stream] = None,
    ):
        """Inits :class:`EMAOptimizer`.

        Parameters
        ----------
        optimizer : torch.optim.Optimizer
            Optimizer to wrap.
        device : torch.device
            Device for EMA parameters.
        decay : float
            Decay factor. Default is ``0.9999``.
        every_n_steps : int
            Apply EMA every N steps. Default is ``1``.
        current_step : int
            Current step. Default is ``0``.
        stream : torch.cuda.Stream, optional
            CUDA stream. Default is ``None``.
        """
        self.optimizer = optimizer
        self.decay = decay
        self.device = device
        self.current_step = current_step
        self.every_n_steps = every_n_steps
        self.save_original_optimizer_state = False

        self.first_iteration = True
        self.rebuild_ema_params = True
        self.stream = stream
        self.thread = None

        self.ema_params: tuple = ()
        self.in_saving_ema_model_context = False

    def all_parameters(self) -> Iterable[torch.Tensor]:
        """Returns an iterator over all parameters."""
        return (param for group in self.param_groups for param in group["params"])

    def step(self, closure=None, **kwargs):  # pylint: disable=unused-argument
        """Performs a single optimization step.

        Parameters
        ----------
        closure : callable, optional
            A closure that reevaluates the model and returns the loss, by default None.
        **kwargs
            Additional parameters.

        Returns
        -------
        float
            Loss.
        """
        self.join()

        if self.first_iteration:
            if any(p.is_cuda for p in self.all_parameters()):
                self.stream = torch.cuda.Stream()

            self.first_iteration = False

        if self.rebuild_ema_params:
            opt_params = list(self.all_parameters())

            self.ema_params = self.ema_params + tuple(
                copy.deepcopy(param.data.detach()).to(self.device) for param in opt_params[len(self.ema_params) :]
            )
            self.rebuild_ema_params = False

        loss = self.optimizer.step(closure)

        if self._should_update_at_step():
            self.update()
        self.current_step = self.current_step + 1
        return loss

    def _should_update_at_step(self) -> bool:
        """Checks if EMA parameters should be updated at current step."""
        return self.current_step % self.every_n_steps == 0

    @torch.no_grad()
    def update(self):
        """Updates EMA parameters"""
        if self.stream is not None:
            self.stream.wait_stream(torch.cuda.current_stream())

        with torch.cuda.stream(self.stream):
            current_model_state = tuple(
                param.data.to(self.device, non_blocking=True) for param in self.all_parameters()
            )

            if self.device.type == "cuda":
                ema_update(self.ema_params, current_model_state, self.decay)

        if self.device.type == "cpu":
            self.thread = threading.Thread(
                target=run_ema_update_cpu,
                args=(
                    self.ema_params,
                    current_model_state,
                    self.decay,
                    self.stream,
                ),
            )
            self.thread.start()

    @staticmethod
    def swap_tensors(tensor1, tensor2):
        """Swaps the values of two tensors in-place."""
        tmp = torch.empty_like(tensor1)
        tmp.copy_(tensor1)
        tensor1.copy_(tensor2)
        tensor2.copy_(tmp)

    def switch_main_parameter_weights(self, saving_ema_model: bool = False):
        """Switches the main parameter weights with the EMA weights."""
        self.join()
        self.in_saving_ema_model_context = saving_ema_model
        for param, ema_param in zip(self.all_parameters(), self.ema_params):
            self.swap_tensors(param.data, ema_param)

    @contextlib.contextmanager
    def swap_ema_weights(self, enabled: bool = True):
        """A context manager to in-place swap regular parameters with EMA parameters. It swaps back to the original
        regular parameters on context manager exit.

        Parameters
        ----------
        enabled : bool, optional
            whether the swap should be performed, by default True
        """
        if enabled:
            self.switch_main_parameter_weights()
        try:
            yield
        finally:
            if enabled:
                self.switch_main_parameter_weights()

    def __getattr__(self, name):
        return getattr(self.optimizer, name)

    def join(self):
        """Wait for the EMA update to finish."""
        if self.stream is not None:
            self.stream.synchronize()

        if self.thread is not None:
            self.thread.join()

    def state_dict(self):
        """Return the optimizer state."""
        self.join()

        if self.save_original_optimizer_state:
            return self.optimizer.state_dict()

        # if we are in the context of saving an EMA model, the EMA weights are in the modules' actual weights
        ema_params = self.ema_params if not self.in_saving_ema_model_context else list(self.all_parameters())
        state_dict = {
            "opt": self.optimizer.state_dict(),
            "ema": ema_params,
            "current_step": self.current_step,
            "decay": self.decay,
            "every_n_steps": self.every_n_steps,
        }
        return state_dict

    def load_state_dict(self, state_dict):
        """Load the optimizer state."""
        self.join()
        self.optimizer.load_state_dict(state_dict["opt"])
        self.ema_params = tuple(param.to(self.device) for param in copy.deepcopy(state_dict["ema"]))
        self.current_step = state_dict["current_step"]
        self.decay = state_dict["decay"]
        self.every_n_steps = state_dict["every_n_steps"]
        self.rebuild_ema_params = False

    def add_param_group(self, param_group):
        """Add a param group to the :class:`Optimizer` s `param_groups`."""
        self.optimizer.add_param_group(param_group)
        self.rebuild_ema_params = True
